Package: betterChromVAR
Title: Improved ChromVAR (Chromatin Variation Across Regions)
Description: 
  A much faster analytical implementation of chromVAR, with additional 
  features, used to infer TF activity from (bulk or single-cell) ATAC-seq data and 
  motif annotations (or binding probabilities). The package also includes the 
  CVnorm normalization method based on the chromVAR logic.
Version: 1.1.9
Authors@R: 
    person("Pierre-Luc", "Germain", email="pierre-luc.germain@hest.ethz.ch", role=c("aut","cre"),
           comment=c(ORCID="0000-0003-3418-4218"))
License: GPL (>= 3)
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
URL: https://github.com/plger/betterChromVAR
BugReports: https://github.com/plger/betterChromVAR/issues
biocViews: Software, ATACSeq, Normalization, Epigenetics, Sequencing
Depends: R (>= 4.1.0), SummarizedExperiment
Imports: BiocParallel, BiocNeighbors, Biostrings, DelayedMatrixStats,
        GenomicRanges, IRanges, Matrix, matrixStats, sparseMatrixStats,
        MatrixGenerics, methods, S4Vectors, stats
Suggests: BiocStyle, knitr, rmarkdown, sessioninfo, testthat
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
git_url: https://git.bioconductor.org/packages/betterChromVAR
git_branch: devel
git_last_commit: 051b316
git_last_commit_date: 2026-08-21
Repository: Bioconductor 3.24
Date/Publication: 2026-08-21
NeedsCompilation: no
Packaged: 2026-08-21 20:36:19 UTC; biocbuild
Author: Pierre-Luc Germain [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3418-4218>)
Maintainer: Pierre-Luc Germain <pierre-luc.germain@hest.ethz.ch>

# Generated by roxygen2: do not edit by hand

export(CVnorm)
export(addGCBias)
export(betterChromVAR)
export(computeBackgrounds)
export(computeDeviationsAnalytic)
export(computeDeviationsFromKNN)
export(computeDeviationsWeighted)
export(computeMotifVariability)
export(getBackgroundBins)
export(getBackgroundKNN)
export(getDummyData)
export(getExpectation)
export(normalizeDevsForSize)
export(sampleBackgroundPeaks)
export(shrinkColumnProps)
exportClasses(bcvBackground)
exportMethods("[")
exportMethods(as.list)
exportMethods(show)
importClassesFrom(Matrix,dgCMatrix)
importClassesFrom(Matrix,dgeMatrix)
importClassesFrom(Matrix,sparseMatrix)
importFrom(BiocNeighbors,findKNN)
importFrom(BiocParallel,MulticoreParam)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpnworkers)
importFrom(Biostrings,getSeq)
importFrom(Biostrings,letterFrequency)
importFrom(DelayedMatrixStats,rowMeans2)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,width)
importFrom(IRanges,IRanges)
importFrom(Matrix,Diagonal)
importFrom(Matrix,Matrix)
importFrom(Matrix,cbind2)
importFrom(Matrix,colSums)
importFrom(Matrix,crossprod)
importFrom(Matrix,drop0)
importFrom(Matrix,kronecker)
importFrom(Matrix,rowMeans)
importFrom(Matrix,rowSums)
importFrom(Matrix,sparseMatrix)
importFrom(Matrix,t)
importFrom(Matrix,tcrossprod)
importFrom(MatrixGenerics,rowMeans2)
importFrom(S4Vectors,metadata)
importFrom(SummarizedExperiment,"assay<-")
importFrom(SummarizedExperiment,"mcols<-")
importFrom(SummarizedExperiment,"rowData<-")
importFrom(SummarizedExperiment,"rowRanges<-")
importFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,assayNames)
importFrom(SummarizedExperiment,colData)
importFrom(SummarizedExperiment,mcols)
importFrom(SummarizedExperiment,rowData)
importFrom(SummarizedExperiment,rowRanges)
importFrom(matrixStats,colMaxs)
importFrom(matrixStats,colMins)
importFrom(matrixStats,rowQuantiles)
importFrom(matrixStats,rowSds)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,validObject)
importFrom(sparseMatrixStats,rowVars)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,dist)
importFrom(stats,dnorm)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,p.adjust)
importFrom(stats,pchisq)
importFrom(stats,poly)
importFrom(stats,qchisq)
importFrom(stats,rnbinom)
importFrom(stats,rnorm)
importFrom(stats,setNames)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
importMethodsFrom(methods,show)

# betterChromVAR

<!-- badges: start -->

<!--
[![GitHub
issues](https://img.shields.io/github/issues/plger/betterChromVAR)](https://github.com/plger/betterChromVAR/issues)
[![GitHub
pulls](https://img.shields.io/github/issues-pr/plger/betterChromVAR)](https://github.com/plger/betterChromVAR/pulls)
[![Bioc release
status](http://www.bioconductor.org/shields/build/release/bioc/betterChromVAR.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/betterChromVAR)
[![Bioc
support](https://bioconductor.org/shields/posts/betterChromVAR.svg)](https://support.bioconductor.org/tag/betterChromVAR)
[![Bioc
history](https://bioconductor.org/shields/years-in-bioc/betterChromVAR.svg)](https://bioconductor.org/packages/release/bioc/html/betterChromVAR.html#since)
[![Bioc
dependencies](https://bioconductor.org/shields/dependencies/release/betterChromVAR.svg)](https://bioconductor.org/packages/release/bioc/html/betterChromVAR.html#since)
-->
[![Bioc devel
status](http://www.bioconductor.org/shields/build/devel/bioc/betterChromVAR.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/betterChromVAR)
[![Bioc last
commit](https://bioconductor.org/shields/lastcommit/devel/bioc/betterChromVAR.svg)](http://bioconductor.org/checkResults/devel/bioc-LATEST/betterChromVAR/)
[![check-bioc](https://github.com/plger/betterChromVAR/actions/workflows/check-bioc.yml/badge.svg)](https://github.com/plger/betterChromVAR/actions/workflows/check-bioc.yml)
[![Codecov test
coverage](https://codecov.io/gh/plger/betterChromVAR/graph/badge.svg)](https://app.codecov.io/gh/plger/betterChromVAR)
<!-- badges: end -->

`betterChromVAR` is a much faster, deterministic implementation of the 
popular [chromVAR](https://greenleaflab.github.io/chromVAR/) (Chromatin 
Variation Across Regions) method for computing cell- (or sample-) wise motif 
deviations from ATAC-seq data and motif annotations. The functions 
additionally have extra options and also include an ATAC-seq normalization 
method based on the general chromVAR logic. See the documentation [here](https://bioconductor.org/packages/devel/bioc/vignettes/betterChromVAR/inst/doc/betterChromVAR.html).

We note that only the components of chromVAR used to compute motif deviations 
were reimplemented here. On a dataset of ~31k cells, ~150k regions, and ~2k 
motifs, the original chromVAR took 138min to run, the ArchR C++ version 78min, 
while betterChromVAR ran in ~2min, with outputs correlating to the original by 
~0.995.

If unfamiliar with chromVAR, we recommend reading up its 
[documentation](https://greenleaflab.github.io/chromVAR/) before using 
betterChromVAR.


## Installation instructions

You can install it from Bioconductor using:

```r
BiocManager::install("betterChromVAR")
```

or install the development version from github with:

```r
BiocManager::install("plger/betterChromVAR")
```

#' CVnorm: chromVAR-inspired ATAC-seq normalization
#' 
#' Corrects ATAC peak counts by removing the effects of technical biases 
#' (GC/accessibility) using the chromVAR background binning approach and an 
#' optional variance-based bias shrinkage (inspired from the qsmooth package) 
#' to preserve group biological signal.
#'
#' @param object A matrix of counts, or a SummarizedExperiment-like object with
#'   an assay named 'counts'.
#' @param bias A vector of length equal to `ncol(object)` specifying the 
#'   per-peak bias (i.e. GC content). If omitted, will try to get it
#'   from `rowData(object)$bias`.
#' @param grouping Optional grouping for the baseline expectation (prevents 
#'   bias toward more abundant groups). This should either be a vector coercible
#'   to factor of length equal to `ncol(object)`, of a character of length 1 
#'   specifying a column of `colData(object)` (if `object` is a 
#'   `SummarizedExperiment`).
#' @param smoothGrouping Optional grouping to determine correction strength. 
#'   If bias is consistent within these groups, correction is reduced. Accepts
#'   the same type of inputs as `grouping`, and by default takes the same 
#'   values.
#' @param shrinkMode The way to perform the group-based shrinkage. With 
#'   `shrinkMode="dampen"` (default), no corrected is applied in bins when the 
#'   bias is entirely explained by groups. `shrinkMode="qsmooth"` instead 
#'   reproduces the logic of the `qsmooth` package: if variance in bias is 
#'   chiefly explained by groups, between group bias will not be corrected, but
#'   within-group differences will be. Using this prior to differential analysis
#'   however leads to increase Type I error rate, and it should therefore not be
#'   used for downstream application.
#' @param toAssay The name of the assay in which to store the corrected data 
#'   (default 'corrected'). Ignored unless `object` is a 
#'   SummarizedExperiment-like object.
#' @param bs Number of bins per dimension (see \code{\link{getBackgroundBins}}).
#' @param w Standard deviation of the Gaussian kernel for bin smoothing.
#' @param Z Logical; whether to return standardized residuals (Z-scores) 
#'   instead of the (default) corrected counts.
#' @param useWidthAdj Whether to adjust for the different width of the regions.
#'   If omitted, will be TRUE if the average absolute difference to the median 
#'   width is greater than 10% of the median width. If TRUE, will adjust for 
#'   `pmax(200L,width)`. If an integer scalar, will adjust for 
#'   `pmax(useWidthAdj,width)`.
#' @param enforceZeros Logical; whether to enforce that zero counts should 
#'   remain zeroes after correction (ignored if `Z=TRUE`).
#' @author Pierre-Luc Germain
#' @references
#'   - Schep A.N., Wu B., Buenrostro J.D., Greenleaf W.J. (2017) chromVAR: 
#'     inferring transcription-factor-associated accessibility from 
#'     single-cell epigenomic data, Nature Methods, doi: 10.1038/nmeth.4401
#'   - Hicks SC, Okrah K, Paulson JN, Quackenbush J, Irizarry RA, Corrado
#'     Bravo H (2018). “Smooth quantile normalization.” Biostatistics 19 (2),
#'     doi: 10.1093/biostatistics/kxx028
#'     
#' @return If `object` is a matrix, then a matrix of corrected counts of the 
#'   same dimensions. If `object` is a SummarizedExperiment-like object, then
#'   the object is returned with an extra assay named based on `toAssay`.
#'   
#' @importFrom SummarizedExperiment assay<- assayNames
#' @importFrom GenomicRanges width
#' @export
#' @examples
#' counts_se <- getDummyData()$counts
#' # if GC content not already in the object, use:
#' # counts_se <- addGCBias(counts_se, genome=YOUR_GENOME)
#' counts_se <- CVnorm(counts_se)
CVnorm <- function(object, bias=NULL, grouping=NULL, smoothGrouping=grouping, 
                   shrinkMode=c("dampen", "qsmooth"), toAssay="corrected",
                   bs=NULL, w=0.1, Z=FALSE, useWidthAdj=NULL,
                   enforceZeros=TRUE){
  
  # input validity
  if(!isFALSE(useWidthAdj) && 
     (!.isSElike(object) || is.null(rowRanges(object))))
    stop("The object does not contain rowRanges.",
         "Either include them, or set `useWidthAdj=FALSE`.")

  flbias <- wi <- NULL  
  if(.isSElike(object)){
    if(is.null(bias)) bias <- rowData(object)$bias
    flbias <- rowData(object)$flbias
    if(is.null(useWidthAdj)){
      wi <- width(object)
      useWidthAdj <- (mean(abs(wi-median(wi)))/median(wi)) > 0.1
    }else{
      stopifnot(isTRUE(useWidthAdj) || 
                  (is.integer(useWidthAdj) && length(useWidthAdj)==1))
      wi <- width(object)
    }
    counts <- assay(object, "counts")
  } else {
    counts <- object
  }
  stopifnot(!is.null(bias) && length(bias) == nrow(counts))
  if(is(counts, "dgeMatrix")) counts <- as.matrix(counts)
  grouping <- .groupingInput(grouping, object)
  smoothGrouping <- .groupingInput(smoothGrouping, object)
  shrinkMode <- match.arg(shrinkMode)

  if(any(rowSums(counts)==0))
    stop("Some peaks have an expectation of zero, most likely because they ",
         "have zero counts. Please remove them.")
  
  # global profile  
  expectation2 <- expectation <- getExpectation(counts, grouping)
  peak_p <- expectation / sum(expectation)
  
  # bias bins
  if(!isFALSE(useWidthAdj)){
    if(isTRUE(useWidthAdj)) useWidthAdj <- 200L
    expectation2 <- useWidthAdj*expectation/pmax(wi, useWidthAdj)
  }
  background <- getBackgroundBins(expectation2, bias=bias, flbias=flbias, w=w, 
                                  bs=bs, verbose=FALSE)
  bin_map <- background@peak2bin
  binBinProbs <- background@binBinProbs
  bin2peakMat <- sparseMatrix(i=bin_map, j=seq_along(expectation), 
                              dims=c(nrow(binBinProbs), length(expectation)))
  
  # Bin-level observed vs expected
  cs <- Matrix::colSums(counts)
  binCounts <- bin2peakMat %*% counts
  bin_p_expected <- as.numeric(bin2peakMat %*% peak_p)
  
  # raw bias factors (B x S)
  smooth_obs_p <- binBinProbs %*% .fastColNorm(binCounts, cs=cs)
  smooth_exp_p <- as.numeric(binBinProbs %*% bin_p_expected)
  bias_factor <- smooth_obs_p / smooth_exp_p
  
  if(!inherits(counts, "sparseMatrix")) bias_factor <- as.matrix(bias_factor)
  
  # variance-based weighting (qsmooth logic)
  if(!is.null(smoothGrouping) && length(unique(smoothGrouping))>1){
    g <- factor(smoothGrouping)
    n_groups <- length(levels(g))
    
    log_R <- log(as.matrix(bias_factor))
    
    # SST (Total Sum of Squares per bin)
    SST <- rowSums((log_R - rowMeans(log_R))^2)
    
    # SSW (Within-group Sum of Squares)
    group_means <- .fastColAgg(log_R, g) %*% Diagonal(x=1/as.numeric(table(g)))
    # Compute means and expand back to sample dimensions
    group_means_mat <- group_means[, as.integer(g)]
    SSW <- Matrix::rowSums((log_R - group_means_mat)^2)
    
    # SSB (Between-group Sum of Squares)
    SSB <- pmax(0, SST - SSW)
    
    # Mean Squares (accounting for degrees of freedom)
    MSW <- SSW / (ncol(counts) - n_groups)
    MSB <- SSB / (n_groups - 1)
    
    weights <- MSW / (MSW + MSB + 1e-12)
    
    if(shrinkMode == "qsmooth"){
      # Subtract the sample-specific noise (log_R - group_mean) 
      # but keep the group-level bias (group_mean) multiplied by (1-w)
      log_R_weighted <- log_R - (1 - weights) * group_means_mat
    } else {
      # Scale down the log-bias factors towards 0 based on the weights
      log_R_weighted <- log_R * weights
    }
    bias_factor <- exp(log_R_weighted)
  }
  
  # peak-level correction
  expected_counts <- (peak_p %*% t(cs)) * bias_factor[bin_map, ]
  
  if(isTRUE(Z)){
    # Pearson residuals
    out <- (counts - expected_counts) / sqrt(expected_counts)
  } else {
    # residuals + global component to keep values in "count-like" scale
    out <- (counts - expected_counts) + (peak_p %*% t(cs))
    if(inherits(out, "sparseMatrix")){
      out@x[out@x < 0] <- 0
    }else{
      out <- as.matrix(out)
      out[out<0] <- 0
    }
    if(isTRUE(enforceZeros)){
      if(inherits(counts, "sparseMatrix")){
        out <- pmin(out,1000*counts)
      }else{
        out[counts==0L] <- 0
      }
    }
  }
  
  if(!inherits(counts, "sparseMatrix")) out <- as.matrix(out)
  
  if(!inherits(object, "SummarizedExperiment")) return(out)
  
  assay(object, toAssay) <- out
  return(object)
}

#' computeDeviationsFromKNNbg
#' 
#' Computes analytical deviations using a nearest-neighbor background matrix 
#' while optionally excluding peaks containing the target motif from its own 
#' background pool.
#' 
#' @param object A SummarizedExperiment or sparse matrix of counts.
#' @param cBg The peak-by-peak sparse kNN matrix, as produced by 
#'  \code{\link{getBackgroundKNN}}.
#' @param annotations Peak annotation (sparse) matrix, with motifs as columns,
#'    or a SummarizedExperiment containing this in the first assay. Values 
#'    should be either logical or between 0 and 1.
#' @param l Lambda parameter determining the weight by which background peaks
#'   containing the foreground motif are scaled in relative importance. Set to 
#'   1 to treat them normally (default), to 0 to exclude them entirely 
#'   (potentially unstable, a small value such as `0.1` is instead recommended).
#' @param chunkSize Number of cells to process simultaneously. Increasing this
#'   will increase speed, but also memory consumption.
#' @param verbose Logical; whether to print progress messages.
#' 
#' @details
#' This method combines the analytic strategy used by `betterChromVAR` with 
#' Ruochi Zhang's approach to use a continuous, multidimensional background 
#' space instead of background bins. If `l<1` it downweighs (multiplying them
#' by `l`) peaks harboring the tested motif from the corresponding motif's 
#' background.
#' 
#' @return A SummarizedExperiment with 'deviations' and 'z' assays. If overall
#'   motif variability and their significance are additionally needed, see 
#'   \code{\link{computeMotifVariability}}.
#' 
#' @importFrom utils txtProgressBar setTxtProgressBar
#' @export
#' @examples
#' attach(getDummyData())
#' bg <- getBackgroundKNN(counts)
#' dev <- computeDeviationsFromKNN(object=counts, cBg=bg,
#'                                 annotations=motifMatches)
#' dev
computeDeviationsFromKNN <- function(object, cBg, annotations, l=1, 
                                     chunkSize=1000, verbose=TRUE){
  stopifnot(nrow(object) == nrow(annotations))
  stopifnot(l>=0 & l<=1)
  
  if(.isSElike(object)){
    counts <- assay(object, "counts")
    depth <- object$depth
    if(is.null(depth)) depth <- Matrix::colSums(counts)
  } else {
    counts <- object
    depth <- Matrix::colSums(counts)
    object <- SummarizedExperiment(list(counts=counts))
  }
  if(.isSElike(annotations))
     annotations <- assay(annotations)
  
  stopifnot(nrow(object)==nrow(annotations))
  stopifnot(nrow(object)==nrow(cBg) && nrow(object)==ncol(cBg))
  nCells <- ncol(counts)
  nMotifs <- ncol(annotations)
  
  # W is M x N
  W <- as(Matrix::t(annotations), "dMatrix")
  k <- Matrix::rowSums(cBg)
  
  if(!is.null(l) && l < 1) {
    if(l > 0){
      if(verbose)
        message("Computing projections with soft motif suppression ",
                "(lambda=", l, ")")
      
      # convert NN to triplet format
      idx_mat <- Matrix::summary(W)
      x <- idx_mat[,3]
      idx_mat <- as.matrix(idx_mat[,1:2])
      
      # S[m, p] is the number of peak p's neighbors that contain motif m
      S <- W %*% Matrix::t(cBg)
      
      motifNeighbors <- S[idx_mat]

      # non-motif containing NNs (denominator)
      denom <- (k[idx_mat[,2]] - motifNeighbors) + (l * motifNeighbors)
      denom[denom <= 0] <- 1 
      
      # regularized weight matrix
      W2 <- Matrix::sparseMatrix(
        i = idx_mat[,1],
        j = idx_mat[,2],
        x = x/denom,
        dims = c(nMotifs, nrow(object))
      )
      
      # apply the soft penalty on motif-containing NNs and compute projections
      WA_raw <- W2 %*% cBg
      WA_corrected <- WA_raw - ((1 - l) * (WA_raw * W))
      rm(WA_raw, idx_mat, W2)
      
    } else {
      if(verbose)
        message("Computing associative projections with strict motif exclusion")
      
      # Calculate fraction of neighbors containing the motif for each peak (M x N)
      denom <- 1 - as.matrix((W %*% Matrix::t(cBg)) %*% Matrix::Diagonal(x=1/k))
      denom[denom <= 0] <- 1 
      
      # Compute background mapping and mask out peaks containing the motif
      W2 <- W %*% Matrix::Diagonal(x = 1/k)
      WA_corrected <- ((W2 / denom) %*% cBg) * (1 - W)
      rm(W2)
    }
    
  } else {
    #if(verbose) message("Computing associative projections")
    WA_corrected <- (W %*% Matrix::Diagonal(x = 1/k)) %*% cBg
  }
  
  if(verbose) message("Computing motif expectations and variances")
  E_motif <- WA_corrected %*% counts 
  Term1 <- WA_corrected %*% counts^2
  rm(WA_corrected)
  gc()
  
  # Variance Term 2: lazy chunking
  Term2 <- matrix(0, nrow = nMotifs, ncol = nCells)
  chunks <- split(seq_len(nCells), ceiling(seq_len(nCells) / chunkSize))
  
  cBgNorm <- Matrix::Diagonal(x=1/k)
  
  showPB <- isTRUE(verbose) && length(chunks) > 1
  if(showPB) pb <- txtProgressBar(min=0L, max=length(chunks), style=3)
  
  for (i in seq_along(chunks)) {
    idx <- chunks[[i]]
    X_chunk <- counts[, idx, drop = FALSE]
    
    AX_avg <- cBgNorm %*% (cBg %*% X_chunk)
    Term2[, idx] <- as.matrix(W %*% (AX_avg^2))
    
    if(showPB) setTxtProgressBar(pb, i)
  }
  if(showPB) close(pb)
  
  V_motif <- as.matrix(Term1) - Term2
  V_motif[V_motif < 0] <- 0 
  
  if(verbose) message("Calculating final deviations and Z-scores")
  observed_motif <- as.matrix(W %*% counts)
  numerator <- observed_motif - as.matrix(E_motif)
  
  z <- numerator / sqrt(pmax(1e-12, V_motif))
  
  peak_means <- Matrix::rowMeans(counts)
  globalMotifAvg <- as.vector(W %*% peak_means)
  sf <- depth / sum(peak_means)
  
  deviations <- numerator / outer(globalMotifAvg, sf)
  
  o <- SummarizedExperiment(list(deviations=deviations, z=z),
                            colData=colData(object), metadata=metadata(object))
  rowData(o)$N <- Matrix::colSums(annotations)
  
  return(o)
}

set.seed(123)
attach(getDummyData())
counts$groups <- factor(rep(LETTERS[1:2], each=5))

checkDevOutput <- function(dev, checkAssayNames=TRUE){
  expect_true(is(dev, "SummarizedExperiment"))
  if(checkAssayNames){
    expect_all_true(c("deviations", "z") %in% assayNames(dev))
  }
  expect_false(any(is.na(assay(dev))))
}

test_that("betterChromVAR with knn background runs", {
  bg <- getBackgroundKNN(counts)
  dev <- computeDeviationsFromKNN(counts, cBg=bg, motifMatches)
  checkDevOutput(dev)
  dev <- computeDeviationsFromKNN(counts, cBg=bg, motifMatches, l=1)
  checkDevOutput(dev)
})

test_that("dgcMatrix supported", {
  assay(counts) <- as(assay(counts), "sparseMatrix")
  bg <- getBackgroundKNN(counts)
  dev <- computeDeviationsFromKNN(counts, cBg=bg, motifMatches)
  checkDevOutput(dev)
})

#' getBackgroundKNN
#' 
#' Computes k-nearest neighbors for each peaks based on the continuous, 
#' multidimensional bias space. This is inpsired by Ruochi Zhang's approach in
#' scPrinter.
#' 
#' @param se A SummarizedExperiment containing a 'counts' assay, or a matrix of 
#'   counts.
#' @param expectation A vector of expectations. If NULL, will use the mean 
#'   counts of `se`
#' @param bias A data.frame of sources of bias (beside `expectation`) to 
#'   consider (one per column), with the same nrow as `se`.
#' @param k Number of nearest neighbors to use.
#' @param pseudo Pseudocount for log transformation.
#' @param weights How to weigh the different bias dimensions. If "none", they
#'   will not be re-weighted. If 'linear' (default), they are weighted by the
#'   absolute Pearson correlation with the over-dispersion. If "poly", by the 
#'   R^2 of a 2nd degree polynomial fit of the over-dispersion.
#' @param ... Passed to \code{\link[BiocNeighbors]{findKNN}}. 
#' 
#' @return A sparse peak-by-peak kNN matrix.
#' @importFrom stats cov lm poly cor setNames
#' @importFrom Matrix sparseMatrix rowMeans
#' @importFrom BiocNeighbors findKNN
#' @importFrom sparseMatrixStats rowVars
#' @importFrom MatrixGenerics rowMeans2
#' @export
#' @examples
#' SE <- getDummyData()$counts
#' bg <- getBackgroundKNN(SE)
getBackgroundKNN <- function(se, expectation=NULL, bias=NULL, k=50,
                                    weights=c("linear","poly","none"), 
                                    pseudo=0.1, ...){
  weights <- match.arg(weights)
  stopifnot(.isSElike(se))
  if(!is.null(bias)){
    stopifnot(is.data.frame(bias) && nrow(bias)==nrow(se))
  }else{
    bias <- data.frame(gc=rowData(se)$bias)
    if(is.null(bias$gc))
      stop("`bias` not provided, and not found in the object.")
  }
  if(is.null(expectation))
    expectation <- MatrixGenerics::rowMeans2(assay(se, "counts"))
  stopifnot(length(expectation)==nrow(se))
  bias <- as.matrix(cbind(macc=log10(expectation+pseudo), bias))

  cv <- cov(bias)
  diag(cv) <- diag(cv) + (pseudo / 1000)
  transMat <- t(forwardsolve(t(chol(cv)), t(bias)))

  if(weights!="none"){
    if(is(assay(se), "dgeMatrix")){
      rvar <- sparseMatrixStats::rowVars(as.matrix(assay(se)))
    }else{
      rvar <- sparseMatrixStats::rowVars(assay(se))
    }
    overd <- log10(rvar/(expectation+1e-6) + pseudo)
    if(weights=="linear"){
      weights <- abs(cor(transMat, overd))
    }else{
      weights <- unlist(apply(transMat, 2, simplify=FALSE, FUN=\(x){
        summary(lm(overd~poly(x, 2, raw=TRUE)))$r.squared
      }))
    }
    transMat <- sweep(transMat, 2, weights/mean(weights), "*")
  }
  
  knn <- BiocNeighbors::findKNN(transMat, k=k, ...)$index

  N <- nrow(transMat)
  bg <- Matrix::sparseMatrix(
    i = rep(seq_len(N), times=k),
    j = as.vector(knn),
    dims = c(N, N)
  )
  attr(bg, "weights") <- setNames(weights, colnames(transMat))
  bg
}

#' computeDeviationsWeighted
#' 
#' A variant of \code{\link{computeDeviationsWeighted}} enabling the 
#' computation of deviations from weighted foreground counts. Specifically, this
#' functions handles the normalization of the difference in magnitude between
#' the (weighted) foreground and background.
#'
#' @param weightedMotifCounts A matrix of weighted counts per motif (rows) and 
#'   sample (columns), or a `SummarizedExperiment` containing this as first 
#'   assay.
#' @param unweightedPeakCounts A matrix of unweighted counts per peak (rows) and 
#'   sample (columns), or a `SummarizedExperiment` containing this as first 
#'   assay.
#' @param annotations Peak annotation (sparse) matrix, with motifs as columns,
#'    or a SummarizedExperiment containing this in the first assay. Values 
#'    should be either logical or between 0 and 1.
#' @param bg Either a \code{\link[betterChromVAR]{bcvBackground-class}} object
#'   as produced by \code{\link{computeBackgrounds}}, or a 
#'   `SummarizedExperiment` of background peak counts (with bias data in 
#'   `rowData`). If missing, will be created based on `unweightedPeakCounts`.
#' @param retSE Logical; whether to return a SummarizedExperiment object.
#' @param ... Passed to \link{getBackgroundBins} (can for instance be used to
#'   pass bias info if not contained in the objects). Ignored if `bg` is a
#'   \code{\link[betterChromVAR]{bcvBackground-class}} object.
#'
#' @returns A SummarizedExperiment (or a list if `retSE=FALSE`). If overall
#'   motif variability and their significance are additionally needed, see 
#'   \code{\link{computeMotifVariability}}.
#' @seealso [computeDeviationsAnalytic()]
#' @export
#' @examples
#' attach(getDummyData())
#' # if GC content not already in the object, use:
#' # counts <- addGCBias(counts, genome=YOUR_GENOME)
#' # For the purpose of this example, we'll use standard (unweighted counts),
#' # although at this step we'd compute counts weighted in the desired fashion:
#' motifCounts <- Matrix::t(motifMatches) %*% assay(counts)
#' dev1 <- computeDeviationsWeighted(motifCounts, counts, motifMatches)
#' dev1
#' # in this case, the results are identical to :
#' dev2 <- betterChromVAR(counts, motifMatches)
#' stopifnot(identical(assays(dev1), assays(dev2)))
computeDeviationsWeighted <- function(weightedMotifCounts, unweightedPeakCounts,
                                      annotations, bg=NULL, retSE=TRUE, ...){
  stopifnot(ncol(weightedMotifCounts)==ncol(unweightedPeakCounts))
  stopifnot(nrow(weightedMotifCounts)==ncol(annotations))
  stopifnot(nrow(unweightedPeakCounts)==nrow(annotations))
  if(is.null(bg)){
    bg <- computeBackgrounds(unweightedPeakCounts,
                             getBackgroundBins(unweightedPeakCounts, ...))
  }else if(.isSElike(bg)){
    stopifnot(nrow(bg)==nrow(annotations) && 
                ncol(bg)==ncol(weightedMotifCounts))
    bg <- computeBackgrounds(bg, getBackgroundBins(bg, ...))
  }else{
    if(length(bg@depth)==0 || length(bg@expectation)==0)
      stop("Incomplete background; please run computeBackgrounds() first.")
    if( (length(bg@depth)!=ncol(weightedMotifCounts)) ||
        (length(bg@expectation)!=nrow(unweightedPeakCounts)) )
      stop("The `background` object does not match the other objects' dimensions.")
  }
  
  CD <- motifCD <- depth <- NULL
  if( inherits(annotations, "SummarizedExperiment") ){
    motifCD <- colData(annotations)
    annotations <- assay(annotations)
  }
  .checkAnnotations(annotations)
  
  if(.isSElike(weightedMotifCounts)){
    CD <- colData(weightedMotifCounts)
    weightedMotifCounts <- assay(weightedMotifCounts)
  }
  if(.isSElike(unweightedPeakCounts)){
    if(is.null(CD)) CD <- colData(unweightedPeakCounts)
    if(is.numeric(unweightedPeakCounts$depth)){
      message("Using pre-computed unweightedPeakCounts$depth")
      depth <- unweightedPeakCounts$depth
    }
    unweightedPeakCounts <- assay(unweightedPeakCounts)
  }
  if(is.null(depth)) depth <- Matrix::colSums(unweightedPeakCounts)
  
  binMap <- bg@peak2bin
  bin2peakMat <- sparseMatrix(i = binMap, j = seq_along(binMap), 
                              dims = c(nrow(bg@binBinProbs), length(binMap)))
  motifBinCounts <- Matrix::t(annotations) %*% Matrix::t(bin2peakMat)
  motif_bg_exp_unweighted <- as.matrix(motifBinCounts %*% bg@E)
  unwMoCounts <- as.matrix(Matrix::t(annotations) %*% unweightedPeakCounts)
  unwMoCounts <- Matrix::rowMeans(.fastColNorm(unwMoCounts))*sum(unwMoCounts)
  fg <- sum(weightedMotifCounts)*Matrix::rowMeans(.fastColNorm(weightedMotifCounts))
  motif_sf <- fg / unwMoCounts
  motif_bg_exp <- motif_bg_exp_unweighted * motif_sf
  
  numerator <- weightedMotifCounts - motif_bg_exp
  
  bg_variance_scaled <- as.matrix(motifBinCounts %*% bg@V) * (motif_sf^2)
  z <- numerator / sqrt(pmax(0, bg_variance_scaled))
  globalMotifAvg <- as.vector(Matrix::crossprod(annotations, 
                                                bg@expectation))
  sf <- bg@depth/sum(bg@expectation)
  deviations <- numerator/outer(globalMotifAvg, sf)
  a <- lapply(list(deviations=deviations, z=z), \(x){
    if(is(x, "dgeMatrix")) x <- as.matrix(x)
    x
  })
  
  if(!isTRUE(retSE)) return(a)
    
  d <- data.frame(N=Matrix::colSums(annotations))
  
  .packageDevSE(a, SummarizedExperiment(list(), colData=CD),motifCD, d)
}

#' computeDeviationsAnalytic
#'
#' @param object A SummarizedExperiment (or SingleCellExperiment) with an assay
#'    'counts', and with a 'bias' column in `rowData(object)`. Note that the 
#'    regions should have similar widths.
#' @param annotations Peak annotation (sparse) matrix, with motifs as columns,
#'    or a SummarizedExperiment containing this in the first assay. Values 
#'    should be either logical or between 0 and 1.
#' @param background A `bcvBackground` object with bins*samples slots filled, 
#'   as produced by \code{\link{computeBackgrounds}}.
#' @param verbose Logical; whether to output progress messages.
#' @param retSE Logical; whether to return a SummarizedExperiment object.
#' @param compute What to compute. Defaults to everything: deviations, z and 
#'   motif variability.
#' @param denominator The type of denominator to use for the deviations.
#'   Either 'global' (default), i.e. the global expectation (same as the 
#'   original chromVAR), 'local' (background expectation of the cell/sample), 
#'   or 'none' (denominator of 1). 'global' (default) is recommended.
#'
#' @returns A SummarizedExperiment (or a list if `retSE=FALSE`).
#' @importFrom Matrix t crossprod tcrossprod
#' @export
#'
#' @examples
#' attach(getDummyData())
#' # if GC content not already in the object, use:
#' # counts <- addGCBias(counts, genome=YOUR_GENOME)
#' 
#' # we fist get the background bins:
#' bg <- getBackgroundBins(counts)
#' # then we compute the backgrounds for each sample:
#' bg <- computeBackgrounds(counts, bg)
#' # then we can compute the deviations:
#' dev <- computeDeviationsAnalytic(counts, bg, motifMatches)
#' dev
computeDeviationsAnalytic <- function(object, background, annotations,
                                      verbose=FALSE, retSE=TRUE,
                                      compute=c("deviations","z","variability"),
                                      denominator=c("global","local","none")){
  # Check input validity
  stopifnot(nrow(object)==nrow(annotations))
  compute <- match.arg(compute, several.ok=TRUE)
  denominator <- match.arg(denominator)
  stopifnot(is(background, "bcvBackground"))

  if(length(background@depth)==0 || length(background@expectation)==0)
    stop("Incomplete background; please run computeBackgrounds() first.")
  
  if( (length(background@depth)!=ncol(object)) ||
      (length(background@expectation)!=nrow(object)) )
    stop("The `background` object does not match the dimensions of `object`.")

  motifCD <- depth <- NULL
  if(.isSElike(object)){
    counts <- assay(object, "counts")
    if(is.numeric(object$depth)){
      if(verbose) message("Using pre-computed object$depth")
      depth <- object$depth
    }
  }else{
    object <- SummarizedExperiment(list(counts=object))
    counts <- assay(object)
  }
  if(!is(counts, "matrix") && !inherits(counts, "Matrix"))
    stop("`object` should be a SummarizedExperiment or SingleCellExperiment,",
         " or a (sparse) matrix of counts.")
  
  if(.isSElike(annotations)){
    motifCD <- colData(annotations)
    annotations <- assay(annotations)
  } 
  .checkAnnotations(annotations)

  if(is.null(depth)) depth <- Matrix::colSums(counts)
  
  binMap <- background@peak2bin
  binBinProbs <- background@binBinProbs

  # sparse mapping from peaks to bins
  bin2peakMat <- sparseMatrix(i=binMap, j=seq_along(binMap), 
                              dims=c(nrow(binBinProbs), length(binMap)))
  
  # motif-level background stats (M x S)
  motifBinCounts <- Matrix::t(bin2peakMat %*% annotations)
  motif_bg_exp <- as.matrix(motifBinCounts %*% background@E)
  observed_motif_sum <- as.matrix(Matrix::crossprod(annotations, counts))
  
  # deviation = (Obs - bgExpect) / bgExpect; z = (Obs-exp)/sdExpect
  deviations <- observed_motif_sum - motif_bg_exp
  
  a <- list()
  if("variability" %in% compute) compute <- union(compute, "z")
  if(any("z" %in% compute)){
    # z = (Obs-exp)/sdExpect
    motif_bg_sd <- sqrt(pmax(0, as.matrix(motifBinCounts %*% background@V)))
    a$z <- deviations / motif_bg_sd
  }
  if("deviations" %in% compute){
    if(denominator=="local"){
      # use the cell's background as expectation
      deviations <- deviations / motif_bg_exp
    }else if(denominator=="global"){
      # global motif expectation (original CV)
      # denom = motif peak counts scaled by the cell's libsize
      globalMotifAvg <- as.vector(Matrix::crossprod(annotations,
                                                    background@expectation))
      sf <- depth / sum(background@expectation)
      deviations <- deviations/outer(globalMotifAvg, sf)
    }
    a$deviations <- deviations
  }
  
  if(!isTRUE(retSE)){
    a$total <- rowSums(observed_motif_sum)
    return(a)
  }
    
  d <- data.frame(N=Matrix::colSums(annotations),
                  total=rowSums(observed_motif_sum))
  
  if("variability" %in% compute){
    d <- cbind(d, computeMotifVariability(a$z))
  }
  
  .packageDevSE(a, object, motifCD, d)
}

.fastColNorm <- function(x, cs=Matrix::colSums(x)){
  x %*% Diagonal(x = 1/cs)
}

.fastColAgg <- function(x, by){
  by <- factor(by)
  stopifnot(length(by)==ncol(x))
  mapMat <- sparseMatrix(
    i = as.integer(by),
    j = seq_along(by),
    dims = c(length(levels(by)), ncol(x)),
    dimnames = list(levels(by), colnames(x))
  )
  tcrossprod(x, mapMat)
}

# Create a 1D Gaussian kernel matrix
#' @importFrom stats median
.diagKernalMatrix <- function(n, sigma, sparsify=TRUE){
  x <- seq_len(n)
  dist_mat <- outer(x, x, "-")
  G <- exp(-(dist_mat^2) / (2 * sigma^2))
  G <- G/rowSums(G)
  if(sparsify && n>10){
    G[which(G<median(G))] <- 0
    G <- as(G, "sparseMatrix")
  }
  G
}

#' getExpectation
#' 
#' Computes expected counts (a glorified rowMeans)
#'
#' @param counts A count matrix, or object inheriting SummarizedExperiment with
#'   a 'counts' assay.
#' @param grouping An optional vector of length equal to `ncol(counts)`,
#'   indicating the grouping of the cells. If provided, cells will be averaged
#'   by group before averaging across groups.
#' @param normalize Logical; whether to normalize data between averaging (but
#'   after grouping). Default TRUE and highly recommended if providing 
#'   `grouping`.
#'
#' @returns A vector of expectation for each row of `counts`
#' @export
#'
#' @importFrom DelayedMatrixStats rowMeans2
#' @examples
#' attach(getDummyData())
#' e <- getExpectation(counts)
getExpectation <- function(counts, grouping=NULL, normalize=TRUE){
  if(.isSElike(counts)) counts <- assay(counts, "counts")
  if(is.null(grouping) || length(unique(grouping))==1)
    return(DelayedMatrixStats::rowMeans2(counts))
  grouping <- factor(grouping)
  stopifnot(length(grouping)==ncol(counts))
  # compute expectation based on an average of group averages
  agcnt <- .fastColAgg(counts, grouping)
  if(is(agcnt, "dgeMatrix")) agcnt <- as.matrix(agcnt)
  if(normalize){
    cs <- Matrix::colSums(agcnt)
    agcnt <- .fastColNorm(agcnt, cs=cs)*median(cs)
    if(is(agcnt, "dgeMatrix")) agcnt <- as.matrix(agcnt)
  }
  DelayedMatrixStats::rowMeans2(agcnt)
}


#' shrinkColumnProps
#' 
#' Empirical Bayes shrinkage of a matrix of counts towards a prior proportion 
#' (by default the mean across columns).
#'
#' @param x A matrix of counts, with features as rows and samples as columns.
#' @param shrinkTo A vector (of length equal to `nrow(x)`) of sampling 
#' probabilities to shrink towards, or a matrix (of the same dimensions as `x`) 
#' of such probabilities. If omitted, will be the weighted mean of the columns'
#' relative frequencies.
#' @param var.theo Logical; whether to use theoretical (i.e. binomial) variances
#'  of the proportions, rather than the observed (weighted) variance.
#'
#' @returns A matrix of the same dimensions as `x` representing the shrunk 
#'  column-wise proportions.
#' @export
#' @importFrom Matrix rowSums
#'
#' @examples
#' # generate a matrix of 5 sampling (with different total counts) of 20 
#' # features based on the same base frequency :
#' baseFreq <- abs(rnorm(20))
#' baseFreq <- baseFreq/sum(baseFreq)
#' mat <- sapply(c(10,20,30,40,50), function(tot){
#'   rpois(length(baseFreq), baseFreq*tot)
#' })
#' # apply shrinkage and confirm that shrunk proportions are better correlated
#' shrunk_mat <- shrinkColumnProps(mat)
#' mean(cor(shrunk_mat))>mean(cor(mat))
shrinkColumnProps <- function(x, shrinkTo=NULL, var.theo=FALSE) {
  cs <- Matrix::colSums(x)
  bigTotal <- sum(cs)
  p <- .fastColNorm(x, cs=cs)
  
  if(is.null(shrinkTo)){
    mu <- Matrix::rowSums(x) / bigTotal
  }else{
    mu <- shrinkTo
  }
  
  if(var.theo){
    # theoretical (i.e. binomial) variances of the proportions
    if(!is.array(mu)){
      pos <- mu*bigTotal
    }else if(is.null(shrinkTo)){
      pos <- x
    }else{
      pos <- mu %*% Diagonal(x=cs)
    }
    v_weighted <- bigTotal*pos*(1-pos)
  }else{
    # weighted row variances
    v_weighted <- Matrix::rowSums(sweep((p - mu)^2, 2, cs, "*")) / bigTotal
  }
  
  # Estimate M (precision parameter)
  # This formula adjusts the observed weighted variance 
  # by subtracting the expected binomial noise
  M <- (mu * (1 - mu) - v_weighted) /
    (v_weighted - (mu * (1 - mu) / mean(cs)))
  
  # Stability Handling
  # If denominator is negative or zero, variance is too low to estimate M
  M[is.na(M) | !is.finite(M) | M <= 0] <- 1000 
  
  # Back-calculate Alpha and Beta Priors
  alpha <- mu * M
  beta <- (1 - mu) * M
  
  # Shrinkage (Posterior Mean)
  # x_ij is the count, n_j is the column sum
  # Result = (x_ij + alpha_i) / (n_j + alpha_i + beta_i)
  
  # Numerator: mat + alpha (vector added to each column)
  # Denominator: n_j (vector) + M (vector) -> requires a matrix
  den_mat <- sweep(matrix(M, nrow = nrow(x), ncol = ncol(x)), 2, cs, "+")
  
  (x + alpha) / den_mat
}


#' Dummy data for testing purposes
#'
#' @param nRegions Number of regions to generate
#' @param nSamples Number of samples to generate
#' @param nMotifs Number of motifs to generate
#'
#' @returns A list with the slots `counts` (a peak counts SummarizedExperiment)
#'   and `matches` (a sparse matrix of binary motif matches per peaks)
#' @importFrom Matrix Matrix
#' @importFrom stats rnorm rnbinom
#' @importFrom GenomicRanges GRanges
#' @importFrom IRanges IRanges
#' @importFrom SummarizedExperiment rowData<-
#' @export
#'
#' @examples
#' out <- getDummyData()
#' (counts <- out$counts)
#' matches <- out$motifMatches
getDummyData <- function(nRegions=500, nSamples=10, nMotifs=5){
  mu <- sample.int(100, nRegions, replace=TRUE)
  counts <- matrix(rnbinom(nRegions * nSamples, mu=rep(mu,nSamples), size=2),
                   nrow=nRegions, ncol=nSamples)
  counts <- counts[which(rowSums(counts)>0),]
  nRegions <- nrow(counts)
  colnames(counts) <- paste0("sample", seq_len(nSamples))
  gr <- GRanges("chr1", IRanges(seq_len(nRegions)*100, width=20))
  counts <- SummarizedExperiment(list(counts=counts), rowRanges=gr)
  rowData(counts)$bias <- pmin(1,pmax(0,rnorm(nRegions, mean=0.5, sd=0.05)))
  matches <- Matrix(
    data=sample(c(0L, 1L), nRegions*nMotifs, replace=TRUE, prob=c(0.85, 0.15)), 
    nrow=nRegions, ncol=nMotifs, sparse=TRUE)
  colnames(matches) <- paste0("motif",seq_len(nMotifs))
  list(counts=counts, motifMatches=matches)
}

#' addGCBias
#' 
#' Add the `bias` column to the object's rowData, containing the regions' 
#' proportion of Gs and Cs.
#'
#' @param object An object inheriting RangedSummarizedExperiment or GRanges.
#' @param genome A BSgenome object or any other genome object supported by 
#'   \code{\link[Biostrings]{getSeq}}.
#'
#' @returns `object` with the GC content in `mcols(object)$bias` (if GRanges) 
#'   or `rowData(object)$bias`.
#' @importFrom Biostrings getSeq letterFrequency
#' @importFrom SummarizedExperiment rowRanges rowRanges<- mcols mcols<-
#' @export
#'
#' @examples
#' # not run:
#' # se <- addGCBias(se, genome)
addGCBias <- function(object, genome){
  if(.isSElike(object)){
    stopifnot(!is.null(rowRanges(object)))
    rowRanges(object) <- addGCBias(rowRanges(object), genome)
    return(object)
  }
  seqs <- Biostrings::getSeq(x=genome, object)
  # same as chromVAR:
  nucfreqs <- letterFrequency(seqs, c("A", "C", "G", "T"))
  gc <- rowSums(nucfreqs[, 2:3]) / rowSums(nucfreqs)
  mcols(object)$bias <- gc
  object
}

.groupingInput <- function(grouping, object, fillNULL=FALSE){
  if(is.null(grouping)){
    if(fillNULL){
      grouping <- rep(factor("all"), ncol(object))
    }else{
      return(NULL)
    }
  }
  if(is.character(grouping) && length(grouping)==1 && 
     .isSElike(object) &&
     grouping %in% colnames(colData(object))){
    grouping <- colData(object)[[grouping]]
  }
  stopifnot(length(grouping)==ncol(object))
  factor(grouping)
}

#' normalizeDevsForSize
#' 
#' Normalizes the z-scores assay of a deviations object to make the scores 
#' comparable across motifs with different number of matches.
#'
#' @param dev A SummarizedExperiment object as produced by 
#'   \code{\link{betterChromVAR}} or \code{\link{computeDeviationsAnalytic}}.
#'
#' @returns The `dev` object with an additional assay named 'norm'.
#' @export
#'
#' @examples
#' attach(getDummyData())
#' dev <- betterChromVAR(counts, motifMatches)
#' dev <- normalizeDevsForSize(dev)
#' dev
normalizeDevsForSize <- function(dev){
  stopifnot(inherits(dev, "SummarizedExperiment"))
  stopifnot("z" %in% assayNames(dev))
  stopifnot(!is.null(rowData(dev)$N))
  N <- rowData(dev)$N
  assay(dev, "norm") <- assay(dev, "z")*sqrt(round(median(N))/N)
  dev
}


.packageDevSE <- function(a, object, motifCD, d){
  if(!is.null(motifCD)) d <- cbind(motifCD, d)
  a <- a[intersect(c("deviations","z"),names(a))]
  SummarizedExperiment(
    assays = a,
    colData = colData(object),
    rowData = d,
    metadata = metadata(object)
  )  
}

.checkAnnotations <- function(annotations){
  stopifnot(length(dim(annotations))==2)
  if(max(annotations) > 1 || min(annotations)<0)
    warning("`annotations` should be either binary or weights from 0 to 1.")
}

.isSElike <- function(object){
  inherits(object, "SummarizedExperiment") ||
    inherits(object, "SingleCellExperiment") ||
    inherits(object, "RangedSummarizedExperiment")
}

#' betterChromVAR
#' 
#' A fast, analytic implementation of `chromVAR`.
#' This is a wrapper around the \code{\link{getBackgroundBins}},
#' \code{\link{computeBackgrounds}}, and \code{\link{computeDeviationsAnalytic}}
#' steps. It additionally allows for multithreading. For more control or 
#' optimization, see the individual steps.
#' 
#' @param object A SummarizedExperiment (or SingleCellExperiment) with an assay
#'    'counts', and with a 'bias' column in `rowData(object)`. Note that the 
#'    regions should have similar widths.
#' @param annotations Peak annotation (sparse) matrix, with motifs as columns,
#'    or a SummarizedExperiment containing this in the first assay. Values 
#'    should be either logical or between 0 and 1.
#' @param grouping An optional factor or vector coercible to a factor indicating
#'   the groupings of the columns of `object`. This is optionally used to 
#'   compute the base expectation such that rare cell types are given as much 
#'   weight as abundant ones. In single-cell data, the grouping can for 
#'   instance be the interaction of samples and cell types. This should either 
#'   be a vector coercible to factor of length equal to `ncol(object)`, or a 
#'   character of length 1 specifying a column of `colData(object)`.
#' @param nthreads Either an integer scalar indicating the number of threads to
#'   use, or a `BiocParallelParam` object.
#' @param verbose Logical; whether to output progress messages (default FALSE).
#' @param ... Passed to \code{\link{getBackgroundBins}}.
#' @author Pierre-Luc Germain
#' 
#' @details
#' Contrarily to the original chromVAR, this function is entirely deterministic,
#' and achieves higher precision and much higher efficiency through two changes:
#' 1) working with expected background sampling mean and variances, rather than
#' actual permutations, and 2) computing expectations and variance at the level
#' of bias bins, instead of in the peak-space. The function additionally 
#' includes experimental bias shrinkage options, the possibility to handle 
#' annotations that are not binary (e.g. probability scores) and a third 
#' bias dimension (fragment length bias, which should be stored in 
#' `rowData(object)$flbias` see \code{\link{getBackgroundBins}} for details).
#' 
#' @references
#'   Schep A.N., Wu B., Buenrostro J.D., Greenleaf W.J. (2017) chromVAR: 
#'   inferring transcription-factor-associated accessibility from 
#'   single-cell epigenomic data, Nature Methods, doi: 10.1038/nmeth.4401
#' 
#' @return A SummarizedExperiment containing the adjusted deviations and 
#'   z-scores for each motif/sample. The rowData additionally contains the 
#'   number of motif matches and their variability.
#' @importFrom SummarizedExperiment SummarizedExperiment assay colData rowData
#' @importFrom S4Vectors metadata
#' @importFrom Matrix crossprod sparseMatrix kronecker Diagonal cbind2 colSums
#' @importFrom BiocParallel bplapply SerialParam MulticoreParam bpnworkers
#' @importFrom stats p.adjust pchisq
#' @importFrom matrixStats rowSds
#' @export
#' @examples
#' attach(getDummyData())
#' # if GC content not already in the object, use:
#' # counts <- addGCBias(counts, genome=YOUR_GENOME)
#' dev <- betterChromVAR(counts, motifMatches)
#' dev
#' # note that this is the exact equivalent of doing:
#' # bg <- getBackgroundBins(counts)
#' # bg <- computeBackgrounds(counts, bg)
#' # dev <- computeDeviationsAnalytic(counts, bg, motifMatches)
betterChromVAR <- function(object, annotations, grouping=NULL, nthreads=NULL,
                           verbose=FALSE, ...){
  
  stopifnot(.isSElike(object))
  stopifnot(nrow(object) == nrow(annotations))
  stopifnot(!is.null(rowData(object)$bias))
  bias <- rowData(object)$bias
  motifCD <- flbias <- NULL
  if(!is.null(rowData(object)$flbias)) flbias <- rowData(object)$flbias
  
  if(.isSElike(annotations)){
    motifCD <- colData(annotations)
    annotations <- assay(annotations)
  } 
  .checkAnnotations(annotations)
  
  grouping <- .groupingInput(grouping, object, TRUE)
  
  if(verbose) message("Preparing bias bins")
  expectation <- getExpectation(object, grouping)
  bg <- getBackgroundBins(expectation, bias=bias, flbias=flbias, 
                          verbose=verbose, ...)
  
  ngroups <- length(levels(grouping))
                    
  if(is.null(nthreads)){
    BPPARAM <- SerialParam(progressbar=(verbose && ngroups>1))
  }else if(is.integer(nthreads) && length(nthreads)==1 && nthreads>0L){
    BPPARAM <- MulticoreParam(nthreads, progressbar=verbose)
  }else{
    if(!inherits(nthreads, "BiocParallelParam"))
      stop("`nthreads` should either be a positive integer, or a ",
           "BiocParallelParam object.")
    BPPARAM <- nthreads
  }
  
  i <- seq_len(ncol(object))
  counts <- assay(object, "counts")
  if((nW <- BiocParallel::bpnworkers(BPPARAM))>1 && ncol(counts) > 100){
    if(verbose) message("Computing backgrounds and deviations")
    chunks <- split(i, cut(i, nW, labels=FALSE))
    res <- bplapply(chunks, BPPARAM=BPPARAM, function(i){
      bg <- computeBackgrounds(counts[,i], bg[,i], expectation=expectation,
                               verbose=FALSE)
      computeDeviationsAnalytic(object[,i], bg, annotations, verbose=FALSE,
                                retSE=FALSE, compute=c("deviations","z"))
    })
    res <- list(deviations=Reduce(cbind2,
                                  lapply(res, function(x) x$deviations)),
                z=Reduce(cbind2, lapply(res, function(x) x$z)),
                total=Reduce("+", lapply(res, function(x) x$total)))
  }else{
    if(verbose) message("Computing backgrounds")
    bg <- computeBackgrounds(counts, bg, expectation=expectation,
                             verbose=verbose)
    if(verbose) message("Computing deviations")
    res <- computeDeviationsAnalytic(object, bg, annotations, verbose=verbose,
                                     retSE=FALSE, compute=c("deviations","z"))
  }
  
  d <- cbind(data.frame(N=colSums(annotations), total=res$total),
             computeMotifVariability(res$z))

  .packageDevSE(res[1:2], object, motifCD, d)
}

#' @importClassesFrom Matrix dgCMatrix dgeMatrix sparseMatrix
setClassUnion("AnyMatrixOrNULL",
              c("matrix", "dgCMatrix", "dgeMatrix", "sparseMatrix","NULL"))
setClassUnion("AnyMatrix",
              c("matrix", "dgCMatrix", "dgeMatrix", "sparseMatrix"))

#' Bin and background data for betterChromVAR (for internal use)
#' @export
setClass("bcvBackground", slots = list(
                                     dims        = "integer",
                                     peak2bin    = "integer",
                                     binDensity  = "integer",
                                     binBinProbs = "AnyMatrix",
                                     E           = "AnyMatrixOrNULL",
                                     V           = "AnyMatrixOrNULL",
                                     expectation = "numeric",
                                     depth       = "numeric"
                                   ))

setValidity("bcvBackground", function(object) {
  errors <- character()
  
  if (length(object@expectation) > 0 &&
      length(object@peak2bin) != length(object@expectation))
    errors <- c(errors, "peak2bin and expectation must have the same length.")

  nb <- length(object@binDensity)
  if(prod(object@dims)!=nb)
    errors <- c(errors, "bin dimensions do not match the actual bins.")

  if(ncol(object@binBinProbs) != nb)
    errors <- c(errors, "binDensity length must match binBinProbs dims.")
  
  if (length(object@E) > 0 && length(object@V) > 0) {
    if (!identical(dim(object@E), dim(object@V))) {
      errors <- c(errors, "E and V must have the same dimensions.")
    }
  }
  
  if (length(object@E) > 0 && nrow(object@E) != length(object@binDensity)) {
    errors <- c(errors, "nrow(E) must match length(binDensity).")
  }
  
  if (length(object@E) > 0 && length(object@depth) > 0 &&
      ncol(object@E) != length(object@depth)) {
    errors <- c(errors, "ncol(E) must match length(depth).")
  }
  
  if (length(errors) == 0) TRUE else errors
})

#' Coerce bcvBackground to a list
#'
#' @rdname bcvBackground-methods
#' @param x A \code{bcvBackground} object.
#' @return A \code{list} containing the slots of the object.
#' @export
setMethod("as.list", "bcvBackground", function(x) {
  list(
    peak2bin    = x@peak2bin,
    binDensity  = x@binDensity,
    binBinProbs = x@binBinProbs,
    E           = x@E,
    V           = x@V,
    expectation = x@expectation,
    depth       = x@depth
  )
})

#' Show a bcvBackground object
#'
#' @rdname bcvBackground-methods
#' @param object A \code{bcvBackground} object.
#' @return Nothing, prints an overview of the object.
#' @importMethodsFrom methods show
#' @export
setMethod("show", "bcvBackground", function(object) {
  cat("bcvBackground object with", length(object@peak2bin), "peaks,\n",
      "split into ", paste(object@dims, collapse="*"), " (",
      length(object@binDensity), ") bins.\n")
  if(!is.null(object@E))
    cat("Background data filled for ", length(object@depth), " samples.")
})

#' Subsetting a bcvBackground
#'
#' @rdname bcvBackground-methods
#' @param x A \code{bcvBackground} object.
#' @param i,j Indices for subsetting (if j is provided, i is ignored).
#' @param ... Additional arguments.
#' @param drop Logical, whether to drop dimensions.
#' @return An \code{bcvBackground} object.
#' @importFrom methods validObject
#' @export
setMethod("[", "bcvBackground", function(x, i, j, ..., drop = TRUE){

  # Handle the case where x[] or x[i] is called instead of x[, j]
  if(missing(j)){
    if(missing(i)) return(x)
    j <- i
  }
  
  if(!is.null(x@E) && ncol(x@E) > 0){
    x@E <- x@E[, j, drop=FALSE]
    x@V <- x@V[, j, drop = FALSE]
  }
  
  if(!is.null(x@depth) && length(x@depth) > 0) x@depth <- x@depth[j]

  validObject(x)
  
  x
})

## ----"install", eval = FALSE--------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE)) {
#       install.packages("BiocManager")
#   }
# 
# BiocManager::install("betterChromVAR")
# 
# ## Check that you have a valid Bioconductor installation
# BiocManager::valid()

## ----loadData-----------------------------------------------------------------
suppressPackageStartupMessages({
  library(SummarizedExperiment)
  library(betterChromVAR)
})
attach(getDummyData())
counts
head(motifMatches)

## ----inspectRowData-----------------------------------------------------------
rowData(counts)

## ----addGCBias, eval=FALSE----------------------------------------------------
# # not run
# counts <- addGCBias(counts, genome=my_genome)

## ----betterChromVAR-----------------------------------------------------------
dev <- betterChromVAR(counts, motifMatches)
dev

## ----variability--------------------------------------------------------------
rowData(dev)

## ----getBackgroundBins--------------------------------------------------------
bg <- getBackgroundBins(counts)
bg

## ----computeBackgrounds-------------------------------------------------------
bg <- computeBackgrounds(counts, bg)
bg

## ----computeDeviations--------------------------------------------------------
dev <- computeDeviationsAnalytic(counts, background=bg, annotations=motifMatches)
dev

## ----getExpectations----------------------------------------------------------
ex <- getExpectation(counts)
bg <- getBackgroundBins(ex, bias=rowData(counts)$bias)
# this is equivalent to bg <- getBackgroundBins(counts)

## ----chunkCompute-------------------------------------------------------------
bg2 <- computeBackgrounds(counts[,1:3], bg, expectation = ex)
dev2 <- computeDeviationsAnalytic(counts[,1:3], bg2, motifMatches)
# this should be identical to what we had run on the whole object:
identical(assay(dev)[,1:3], assay(dev2))

## ----CVnorm-------------------------------------------------------------------
# we assign arbitrary groups to the samples:
counts$group <- rep(LETTERS[1:2], each=5)
# we run the smoothed CVnorm:
counts <- CVnorm(counts, grouping=counts$group)
# (the normal CVnorm could be run by omitting the grouping)
counts

## -----------------------------------------------------------------------------
bg <- getBackgroundKNN(counts)
dev <- computeDeviationsFromKNN(counts, bg, motifMatches)

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

#' computeBackgrounds
#'
#' @param object A SummarizedExperiment (or SingleCellExperiment) with an assay
#'    'counts', or a (sparse) matrix of counts.
#' @param bins A `bcvBackground` object, as produced by 
#'   \code{\link{getBackgroundBins}}.
#' @param grouping An optional factor or vector coercible to a factor indicating
#'   the groupings of the columns of `object`. This is optionally used to 1) 
#'   compute the base expectation such that rare cell types are given as much 
#'   weight as abundant ones, and 2) apply shrinkage (if `shrinkage!="none"`) 
#'   on a per-grouping fashion. In single-cell data, the grouping can for 
#'   instance be the interaction of samples and cell types. (The name of a 
#'   colData column of `object` can also be provided.)
#' @param expectation Optional vector of length equal to `nrow(object)` 
#'   giving the expected counts. If NULL, defaults to mean counts (eventually
#'   grouped, see `grouping` and \code{\link{getExpectation}}).
#' @param shrinkage The method to use to shrink background (i.e. bias) bin 
#'   frequencies. Either "average" (shrinks towards the bin's average across 
#'   cells/samples of the same group), "smooth" (per-sample 2D smoothing over
#'   the bin matrix, somewhat redundant with `w`), or "none" (default).
#' @param sigma Sigma parameter for the 2D smoothing. Ignored unless 
#'   `shrinkage="smooth"`.
#' @param verbose Logical; whether to output progress messages.
#'
#' @returns A `bcvBackground` object with bins*samples slots filled, for use
#'   with \code{\link{computeDeviationsAnalytic}}.
#' @export
#'
#' @examples
#' attach(getDummyData())
#' # if GC content not already in the object, use:
#' # counts <- addGCBias(counts, genome=YOUR_GENOME)
#' 
#' # we fist get the background bins:
#' bg <- getBackgroundBins(counts)
#' # then we can compute the backgrounds for each sample:
#' bg <- computeBackgrounds(counts, bg)
#' # for use in computeDeviationsAnalytic...
computeBackgrounds <- function(object, bins, grouping=NULL, expectation=NULL,
                               shrinkage=c("none", "average", "smooth"),
                               sigma=1, verbose=FALSE){
  # Check input validity
  shrinkage <- match.arg(shrinkage)
  stopifnot(is.null(expectation) || length(expectation)==nrow(object))
  depth <- NULL
  if(.isSElike(object)){
    counts <- assay(object, "counts")
    if(is.numeric(object$depth)){
      if(verbose) message("Using pre-computed object$depth")
      depth <- object$depth
    }
  }else{
    object <- SummarizedExperiment(list(counts=object))
    counts <- assay(object)
  }
  stopifnot(length(dim(counts))==2)
  if(!is(counts, "matrix") && !inherits(counts, "Matrix"))
    stop("`object` should be a SummarizedExperiment or SingleCellExperiment,",
         " or a (sparse) matrix of counts.")
  
  stopifnot(is(bins, "bcvBackground"))
  if(length(bins@peak2bin)!=nrow(object))
    stop("The number of peaks in the `bins` object does not match the rows",
         "of `object`.")
  
  if(length(bins@depth)>0) depth <- bins@depth
  if(is.null(depth)) depth <- colSums(counts)

  if(is.null(expectation)){
    if(length(bins@expectation)==nrow(counts)){
      if(verbose) message("Using pre-computed expectation")
      expectation <- bins@expectation
    }else{
      expectation <- getExpectation(counts, grouping)
    }
  }
  if(any(expectation==0)){
    stop("Some peaks have an expectation of zero, most likely because they ",
         "have zero counts. Please remove them.")
  }
  
  grouping <- .groupingInput(grouping, object)

  bin_map <- bins@peak2bin
  binBinProbs <- bins@binBinProbs
  
  # sparse mapping from peaks to bins
  bin2peakMat <- sparseMatrix(i=bin_map, j=seq_along(expectation), 
                              dims=c(nrow(binBinProbs), length(expectation)))
  
  binCounts <- bin2peakMat %*% counts
  
  if(shrinkage != "none"){
    if(verbose) message("Applying shrinkage")
    il <- split(seq_len(ncol(binCounts)), grouping)
    binCounts <- Reduce(cbind2, lapply(il, function(i){
      binCounts2 <- binCounts[,i]
      cs2 <- depth[i]
      if(shrinkage=="average"){
        # method of moment shrinkage towards per-bin average across cells
        binCounts2 <- shrinkColumnProps(binCounts2)
      }else if(shrinkage=="smooth"){
        # method of moment shrinkage towards cell's 2D-smoothed proportions
        stopifnot(sigma>0)
        G <- .diagKernalMatrix(sqrt(nrow(binCounts2)), sigma=sigma)
        # Create the 2D Kronecker Smoothing Matrix
        G <- Matrix::kronecker(G, G)
        binCounts2 <- shrinkColumnProps(binCounts2,
                                        .fastColNorm(G %*% binCounts2))
      }
      binCounts2 %*% Diagonal(x=cs2)
    }))
    binCounts <- binCounts[,order(unlist(il))]
  }
  
  # bin-level expectations and variances (B x S)
  E <- (binBinProbs %*% binCounts)
  di <- Diagonal(x = 1/pmax(1, bins@binDensity))
  V <- (di %*% (bin2peakMat %*% (counts^2))) - (di %*% E)^2
  V@x[which(V@x<0)] <- 0
  V <- drop0(V)
  V <- binBinProbs %*% (V * bins@binDensity)
  
  if(is(E,"dgeMatrix")) E <- as.matrix(E)
  if(is(V,"dgeMatrix")) V <- as.matrix(V)
  
  bins@E <- E
  bins@V <- V
  bins@expectation <- expectation
  bins@depth <- depth
  validObject(bins)
  
  bins
}

#' getBackgroundBins
#' 
#' Computes chromVAR-like background (i.e. bias) bins, as well as bin-to-bin 
#' selection probabilities needed for \code{\link{betterChromVAR}}.
#' 
#' @param x A SummarizedExperiment containing a 'counts' assay, or a matrix of
#'   counts, or a vector of expected (e.g. mean) counts.
#' @param bias A vector of length equal to `ncol(object)` specifying the 
#'   per-peak bias (i.e. GC content). If omitted, will try to get it
#'   from `rowData(x)$bias`.
#' @param flbias A vector of length equal to `ncol(object)` specifying the 
#'   per-peak fragment length bias (e.g. log10-transformed median length of 
#'   fragments overlapping each region). If omitted, will try to get it
#'   from `rowData(object)$flbias`. If absent, will use 2-dimensional bias bins
#'   (which is already pretty good).
#' @param w Standard deviation of the Gaussian kernel.
#' @param bs Number of bins per dimension. This can be a single integer (total
#'   bins = `bs^2`), or an integer vector of length 2 (if `flbias=NULL`) or 3 
#'   (in which case there are `prod(bs)` total bins). The values specify the 
#'   number of bins for, in order: enrichment, GC and fragment length. By 
#'   default, `bs=50` if `flbias` is not provided (mimicking chromVAR), and
#'   `bs=c(30, 30, 6)` if it is.
#' @param pseudo Optional pseudocount to be added. This should not be needed 
#'   with standard workflows.
#' @param verbose Whether to print processing info.
#' 
#' @details
#' The procedure underlying this function is the same as in 
#' `chromVAR::getBackgroundPeaks`, with the following differences:
#' * Rather than producing a set of background peaks for each input peak, the 
#'   function returns peak-to-bin mappings and bin-to-bin background selection
#'   probabilities, which enables an analytic background computation. It is, as
#'   such, entirely deterministic.
#' * The function supports the optional use of a third bias dimension, provided
#'   through the `flbias` argument, meant for fragment length bias. This is 
#'   still an experimental feature.
#' 
#' @return A `bcvBackground` object, to be used with 
#'   \code{\link{computeBackgrounds}}.
#' @references
#'   Schep A.N., Wu B., Buenrostro J.D., Greenleaf W.J. (2017) chromVAR: 
#'   inferring transcription-factor-associated accessibility from 
#'   single-cell epigenomic data, Nature Methods, doi: 10.1038/nmeth.4401
#'   
#' @importFrom stats cov dnorm dist
#' @importFrom SummarizedExperiment assay rowData
#' @importFrom matrixStats colMins colMaxs
#' @importFrom methods as is new
#' @importFrom Matrix drop0
#' @export
#' @examples
#' counts_se <- getDummyData()$counts
#' background <- getBackgroundBins(counts_se)
getBackgroundBins <- function(x, bias=NULL, flbias=NULL, w=0.1, bs=NULL,
                              pseudo=0, verbose=TRUE){
  if(.isSElike(x)){
    if(is.null(bias)) bias <- rowData(x)$bias
    if(is.null(flbias)) flbias <- rowData(x)$flbias
    x <- Matrix::rowMeans(assay(x, "counts"))
  }else if(is.matrix(x) || is(x, "Matrix")){
    x <- Matrix::rowMeans(x)
  }
  if(is.null(bias)) stop("`bias` not provided, and not found in the object.")
  stopifnot(length(bias)==nrow(x))
  
  stopifnot("Some peaks have zero counts; please remove them."=sum(x==0)==0)
  
  if(!is.null(flbias)){
    if(is.null(bs)) bs <- c(30L, 30L, 6L)
    stopifnot(length(bs)==3)
    stopifnot(length(flbias)==length(bias))
  }else{
    if(is.null(bs)) bs <- 50
    if(length(bs)==1) bs <- c(bs,bs)
    stopifnot(length(bs)==2)
  }
  bs <- as.integer(bs)
  stopifnot(all(bs>=1))
  if(verbose) 
    message("Creating ", paste(bs,collapse="*"),"=",prod(bs)," bias bins ",
            "and computing their sampling distances")
  
  # Mahalanobis transformation
  norm_mat <- cbind(log10(x+pseudo), bias)
  if(!is.null(flbias)) norm_mat <- cbind(norm_mat, flbias)
  cv <- cov(norm_mat)
  diag(cv) <- diag(cv) + pseudo/1000
  transMat <- t(forwardsolve(t(chol(cv)), t(norm_mat)))
  
  # Calculate min/max for the bin boundaries
  minCoords <- colMins(transMat)
  maxCoords <- colMaxs(transMat)
  range_coords <- maxCoords - minCoords
  
  # Map peak coordinates to bins
  idx1 <- round((transMat[, 1] - minCoords[1]) / 
                  range_coords[1] * (bs[1] - 1)) + 1
  idx2 <- round((transMat[, 2] - minCoords[2]) /
                  range_coords[2] * (bs[2] - 1)) + 1
  # Ensure indices fall within [1, bs]
  idx1 <- pmax(1, pmin(bs[1], idx1))
  idx2 <- pmax(1, pmin(bs[2], idx2))
  
  if(!is.null(flbias)){
    idx3 <- round((transMat[, 3] - minCoords[3]) /
                    range_coords[3] * (bs[3] - 1)) + 1
    idx3 <- pmax(1, pmin(bs[3], idx3))
  }
  
  # Linearize index
  peak2bin <- idx1 + (idx2 - 1) * bs[1]
  if(!is.null(flbias)){
    peak2bin <- peak2bin  + (idx3 - 1) * (bs[1] * bs[2])
  }
  
  binDensity <- tabulate(peak2bin, nbins = prod(bs))
  
  # bin center grid (for distance calculation)
  grid_args <- lapply(seq_along(minCoords), function(i) {
    seq(minCoords[i], maxCoords[i], length.out = bs[i])
  })
  bin_data <- do.call(expand.grid, grid_args)

  # bin-to-bin probability matrix
  W <- dnorm(as.matrix(dist(bin_data)), 0, w)
  
  normalizer <- as.vector(W %*% binDensity)
  # Avoid division by zero for empty regions
  normalizer[normalizer < 1e-9] <- 1
  binBinProbs <- W/normalizer
  
  binBinProbs[binBinProbs < 1e-6] <- 0
  binBinProbs <- as(binBinProbs, "sparseMatrix")
    
  new("bcvBackground",
      dims = bs,
      peak2bin = as.integer(peak2bin), 
      binDensity = binDensity, 
      binBinProbs = binBinProbs,
      E = NULL,
      V = NULL,
      expectation = numeric(), 
      depth = integer())
}


#' sampleBackgroundPeaks
#' 
#' Given a background generated by \code{\link{getBackgroundBins}}, samples 
#' background peaks for each input peak.
#'
#' @param background A `bcvBackground` produced by 
#'   \code{\link{getBackgroundBins}}.
#' @param niterations Number of background peaks to sample for each target peak. 
#'
#' @return A peaks x niterations matrix of integers representing the indices of
#'   the sampled background peaks.
#'   
#' @details
#' This function is not used by \code{\link{betterChromVAR}}, which is 
#' deterministic, but for other applications requiring an outputs similar to 
#' that of the original \code{\link[chromVAR]{getBackgroundPeaks}}.
#' 
#' @export
#'
#' @examples
#' counts_se <- getDummyData()$counts
#' background <- getBackgroundBins(counts_se)
#' bg_peaks <- sampleBackgroundPeaks(background, niterations=20)
sampleBackgroundPeaks <- function(background, niterations=50){
  stopifnot(is(background, "bcvBackground"))
  peak2bin <- background@peak2bin
  binBinProbs <- background@binBinProbs
  density <- background@binDensity
  n_bins <- nrow(binBinProbs)

  out <- matrix(0L, nrow=length(peak2bin), ncol=niterations)
  
  peaks_in_bin <- split(seq_along(peak2bin), peak2bin)
  
  for (b in names(peaks_in_bin)) {
    target_peaks <- peaks_in_bin[[b]]
    b <- as.integer(b)
    n_targets <- length(target_peaks)
    n_samples <- n_targets * niterations
    
    # Prob of selecting candidate bin `j` for target bin `b` is proportional to:
    # (Probability of selecting a specific peak in `j`) * (Total peaks in `j`)
    b_probs <- as.numeric(binBinProbs[b, ]) * density
    if(sum(b_probs) == 0) b_probs[b] <- 1
    
    # sample bins
    sampled_bins <- sample(n_bins, size=n_samples, replace=TRUE, prob=b_probs)
    
    # convert sampled bins to sampled peaks
    sampled_counts <- tabulate(sampled_bins, nbins = n_bins)
    active_candidate_bins <- which(sampled_counts > 0)
    
    sampled_peaks <- integer(n_samples)
    for (j in active_candidate_bins) {
      cand_peaks <- peaks_in_bin[[as.character(j)]]
      # handle the single-peak bin edge case to avoid sampling from 1:x.
      if(length(cand_peaks) == 1){
        sampled_peaks[sampled_bins == j] <- cand_peaks
      }else{
        sampled_peaks[sampled_bins == j] <- 
          sample(cand_peaks, size=sampled_counts[j], replace=TRUE)
      }
    }
    
    # assign to the target rows in the output matrix
    out[target_peaks, ] <- matrix(sampled_peaks, ncol=niterations)
  }
  
  return(out)
}

library(BiocParallel)
library(Matrix)
set.seed(123)
attach(getDummyData())
counts$groups <- factor(rep(LETTERS[1:2], each=5))

checkDevOutput <- function(dev, checkAssayNames=TRUE){
  expect_true(is(dev, "SummarizedExperiment"))
  if(checkAssayNames){
    expect_all_true(c("deviations", "z") %in% assayNames(dev))
  }
  expect_false(any(is.na(assay(dev))))
}

test_that("betterChromVAR runs", {
  dev <- betterChromVAR(counts, motifMatches)
  checkDevOutput(dev)
})

test_that("dgcMatrix supported", {
  assay(counts) <- as(assay(counts), "sparseMatrix")
  dev <- betterChromVAR(counts, motifMatches)
  checkDevOutput(dev)
})


bg <- getBackgroundBins(counts)

test_that("shrinkage works", {
  bg <- computeBackgrounds(counts, bg, grouping=counts$groups,
                           shrinkage="average")
  dev <- computeDeviationsAnalytic(counts, bg, motifMatches)
  checkDevOutput(dev)
  bg <- computeBackgrounds(counts, bg, grouping=counts$groups,
                           shrinkage="smooth")
  dev <- computeDeviationsAnalytic(counts, bg, motifMatches)
  checkDevOutput(dev)
})

test_that("multithreading works", {
  bp <- SnowParam(2)
  dev <- betterChromVAR(counts, motifMatches, grouping=counts$groups,
                        nthreads=bp)
  checkDevOutput(dev)
})

test_that("CVnorm works", {
  counts <- CVnorm(counts, grouping=counts$groups)
  checkDevOutput(counts, FALSE)
  expect_true("corrected" %in% assayNames(counts))
})

test_that("Fragment length bias works", {
  rowData(counts)$flbias <- pmax(rnorm(nrow(counts), 2.5, 0.25),0.5)
  background <- getBackgroundBins(counts, bs=c(10,10,4))
  expect_equal(nrow(background@binBinProbs), (10*10*4))
})

test_that("Sampling background peaks works", {
  background <- getBackgroundBins(counts)
  bg_peaks <- sampleBackgroundPeaks(background, niterations=10)
  expect_all_true(dim(bg_peaks)==c(nrow(counts), 10L))
})


set.seed(123)
z <- matrix(rnorm(mean=rnorm(10), sd=runif(10, max=2), 1000), nrow=10)

compareVarDat <- function(x,y,isP=FALSE){
  if(isP) return(1-cor(x,y))
  mean(abs(x-y)/x)
}

test_that("computeMotifVariability works", {
  v0 <- computeMotifVariability(z, confInt = 0.6)
  v1 <- computeMotifVariability(z, confInt = 0.6, method="normal")
  v2 <- computeMotifVariability(z, confInt = 0.6, method="bootstrap")
  diff <- vapply(colnames(v0)[1:4], FUN.VALUE=numeric(2), \(x){
    c(compareVarDat(v0[[x]], v1[[x]], grepl("pval", x)),
      compareVarDat(v0[[x]], v2[[x]], grepl("pval", x)))
  })
  expect_all_true(as.numeric(diff)<0.05)
})

#' computeMotifVariability
#'
#' @param z A matrix of z-scores (with motifs as rows), or a 
#'   `SummaziedExperiment` object with such a matrix in assay `z` (as produced
#'   e.g. by \code{\link{computeDeviationsAnalytic}}).
#' @param confInt The the confidence interval (a numeric scalar between 0 and 
#'   1, default 0.95).
#' @param method The method used to compute the confidence interval. 'normal'
#'   computes it analytically, assuming that the z-scores are normally 
#'   distributed. 'bonett' (default) adjusts this analytic estimate for 
#'   kurtosis (based on Bonett, Computational Statistics & Data Analysis, 2006).
#'   'bootstrap' uses bootstrapping, which does not scale very well.
#' @param n The number of bootstrap samples. Ignored unless 
#'   `method="bootstrap"`.
#'
#' @returns A data.frame containing the variability and confidence interval 
#'   around it, as well as significance, for each motif. If `z` is a 
#'   `SummarizedExperiment`, the data.frame will be stored in `rowData(z)`.
#' @export
#' @importFrom stats qchisq
#' @importFrom matrixStats rowSds rowQuantiles
#' @importFrom MatrixGenerics rowMeans2
#' @examples
#' # we generate random z-scores:
#' z <- matrix(rnorm(mean=rnorm(10), sd=runif(10, max=2), 200), nrow=10)
#' var <- computeMotifVariability
computeMotifVariability <- function(z, confInt=0.95, n=100,
                                    method=c("bonett", "normal", "bootstrap")){
  method <- match.arg(method)
  stopifnot(length(confInt)==1 && confInt>0 && confInt<1)

  if(.isSElike(z)){
    out <- computeMotifVariability(assay(z, "z"), confInt, method)
    rowData(z) <- cbind(
      rowData(z)[,setdiff(colnames(rowData(z)),colnames(out)),drop=FALSE], out)
    return(z)
  }
  
  if(method=="bonett") return(computeBonettVariability(z, confInt))
  if(method=="bootstrap") return(computeBootstrapVariability(z, confInt, n=n))
  
  df <- ncol(z)-1L
  variability <- matrixStats::rowSds(as.matrix(z))
  vars <- variability^2
  
  alpha2 <- (1-confInt)/2
  chi_ci <- stats::qchisq(c(1-alpha2, alpha2), df=df)
  
  pval <- stats::pchisq(df*vars, df=df, lower.tail=FALSE)
  
  data.frame(
    variability = variability,
    var.lower = sqrt(df*vars/chi_ci[1]),
    var.upper = sqrt(df*vars/chi_ci[2]),
    var.pval=pval,
    var.adjPval=p.adjust(p=pval, method="BH"),
    row.names = rownames(z)
  )
}


# Compute variability using Bonett's Method (Kurtosis-adjusted variability)
# 
# @param z_scores A dense or sparse M x C matrix of analytical Z-scores.
# @param confInt The confidence interval (default 0.95).
# @return A data.frame with variability (SD), kurtosis, and confidence bounds.
computeBonettVariability <- function(z, confInt=0.95){
  z <- as.matrix(z)
  nC <- ncol(z)
  SD <- matrixStats::rowSds(z)
  vars <- SD^2
  z <- z - rowMeans2(z)
  
  # Calculate kurtosis
  m2 <- rowMeans2(z^2)
  kurtosis <- ifelse(m2 == 0, 3, rowMeans2(z^4) / (m2^2))
  
  # Bonett's CI calculation
  # Get the standard normal critical value
  alpha <- 1 - confInt
  z_crit <- stats::qnorm(1 - alpha / 2)
  
  # Bonett's standard error of the log-variance
  se_ln_var <- sqrt( (kurtosis - (nC-3)/nC) / (nC-1) )
  # Bonett's small-sample adjustment constant 'c'
  # (approaches 1 for large nC)
  c <- nC / (nC - z_crit)
  ln_var <- log(c * pmax(vars, 1e-16))
  
  # Compute CI for the Standard Deviation 
  ci_lower <- exp(0.5 * (ln_var - z_crit * se_ln_var))
  ci_upper <- exp(0.5 * (ln_var + z_crit * se_ln_var))
  
  # Correct the bounds for motifs with absolutely zero variance
  ci_lower[which(vars==0)] <- 0
  ci_upper[which(vars==0)] <- 0
  
  pval <- stats::pnorm(ln_var/se_ln_var, lower.tail=FALSE)
  
  data.frame(
    variability=SD,
    var.lower=ci_lower,
    var.upper=ci_upper,
    var.pval=pval,
    var.adjPval=p.adjust(p=pval, method="BH"),
    row.names=rownames(z)
  )
}

computeBootstrapVariability <- function(z, confInt=0.95, n=100){
  z <- as.matrix(z)
  bsds <- vapply(seq_len(n), FUN.VALUE=numeric(nrow(z)), FUN=\(i){
    idx <- sample.int(ncol(z), size=ncol(z), replace=TRUE)
    matrixStats::rowSds(z[, idx])
  })
  alpha2 <- (1-confInt)/2
  CI <- matrixStats::rowQuantiles(bsds, probs=c(alpha2, 1-alpha2))
  pval <- (rowSums(bsds <= 1)+1) / (ncol(bsds)+1)
  data.frame(
    variability=matrixStats::rowSds(z),
    var.lower=CI[,1],
    var.upper=CI[,2],
    var.pval=pval,
    var.adjPval=p.adjust(p=pval, method="BH"),
    row.names=rownames(z)
  )
}

---
title: "Introduction to betterChromVAR"
author: 
  - name: Pierre-Luc Germain
    affiliation:
    - ETH Zurich, Switzerland
    email: pierre-luc.germain@hest.ethz.ch
date: "`r doc_date()`"
output: 
  BiocStyle::html_document:
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
    code_folding: show
bibliography: ../inst/REFERENCES.bib
package: "`r pkg_ver('betterChromVAR')`"
vignette: >
  %\VignetteIndexEntry{Introduction to betterChromVAR}
  %\VignetteEncoding{UTF-8}  
  %\VignetteEngine{knitr::rmarkdown}
---

# Introduction

The `r Biocpkg("chromVAR")` R package was originally developed by Alicia
Schep and colleagues from the Greenleaf lab [@schep2017chromvar].
Although originally designed for single-cell ATAC-seq data, it has been
shown to be highly sensitive for bulk data as well
[@gerbaldo2024identification]. The aim of the method is to infer, based
on the accessibility of motif matches, the relative activity of
transcription factors (TFs) in each sample or cell, adjusting for
technical biases (GC content and enrichment bias). It is recommended
that you read the `r Biocpkg("chromVAR")` documentation before using
this package.

`r Biocpkg("betterChromVAR")` is first and foremost a considerably
faster, analytical re-implementation of the original method (it is also
considerably faster than the C++ reimplementation in ArchR
[@granja2021archr]. Contrarily to the original chromVAR, it is entirely
deterministic, and achieves much higher efficiency by replacing permutations 
with an analytical solution and computing expectations and variance at the 
level of bias bins, instead of in the peak-space.

In addition, `r Biocpkg("betterChromVAR")` includes a few additions,
such as simpler weighted expectations, bias shrinkage, and an ATAC-seq
normalization method based on the chromVAR logic. Importantly, however,
not all functionalities of chromVAR have been reimplemented here:
`r Biocpkg("betterChromVAR")` chiefly focuses on the key task of
efficiently computing motif deviations.

# Installing `betterChromVAR`

`r Biocpkg("betterChromVAR")` is a `R` package available via the
[Bioconductor](http://bioconductor.org) repository. It can be installed
using the following commands in your `R` session:

```{r "install", eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE)) {
      install.packages("BiocManager")
  }

BiocManager::install("betterChromVAR")

## Check that you have a valid Bioconductor installation
BiocManager::valid()
```

# Computing motif deviations with betterChromVAR

`r Biocpkg("betterChromVAR")` takes two primary inputs: 1) the counts in
peaks across cells or samples, and 2) an annotation of which TFs/motifs
match which peak[^1]. The peak counts should be provided as a
`RangedSummarizedExperiment` object[^2], while the annotation can either
be in that format too or provided as a (sparse) matrix.

[^1]: The annotation can be binary or probabilistic (i.e. from 0 to 1).
    If using motif matches, however, it is recommended to input binary
    matches, as the magnitude of motif scores is generally poorly
    correlated to actual binding.

[^2]: See `r Biocpkg("SummarizedExperiment")` if you're not familiar
    with those

There are multiple ways of generating your peak counts; the original
`r Biocpkg("chromVAR")` package includes such a function
(`getCounts()`), and the `r Githubpkg("ETHZ-INS/epiwraps", "epiwraps")`
package has some with more options (functions `peakCountsFromBAM()` and
`peakCountsFromFrags()`). Similarly, the `r Biocpkg("motifmatchr")`
package can be used to generate the motif matching annotation. An
important consideration is that the peaks or regions used for the
purpose of this analysis should have similar widths. It is thus highly
recommended, before generating the count and annotation matrices, to
resize your regions (e.g. using
`peaks <- resize(peaks, width=300, fix="center")`). The exact size can
be something not too large (otherwise the presence/absence of a motif
becomes meaningless) and ideally close to the median size of your
original peaks. In addition, is it advisable to restrict your peaks to
those that are on standard chromosomes[^3].

[^3]: See `keepStandardChromosomes()` from the
    `r Biocpkg("GenomeInfoDb")` package

Here we'll use dummy data as example:

```{r loadData}
suppressPackageStartupMessages({
  library(SummarizedExperiment)
  library(betterChromVAR)
})
attach(getDummyData())
counts
head(motifMatches)
```

In this case, we can see that `rowData(counts)` already has a `bias`
column indicating the GC content of the regions:

```{r inspectRowData}
rowData(counts)
```

Had this not been the case, we would first need to add this using:

```{r addGCBias, eval=FALSE}
# not run
counts <- addGCBias(counts, genome=my_genome)
```

where `my_genome` is a `r Biocpkg("BSgenome")` object or similar (e.g.
an `FaFile`[^4]).

[^4]: See the `r Biocpkg("Rsamtools")` package.

Once we have this, we can launch the computation of the deviations :

```{r betterChromVAR}
dev <- betterChromVAR(counts, motifMatches)
dev
```

The resulting `dev` object is a SummarizedExperiment with the same
columns (and colData) as the original `counts` object, but with the
motifs as rows (instead of the original peaks). The values can be
interpreted as the relative activity, across samples or cells, of the
corresponding TFs[^5]. When the function is used with default arguments,
the results are virtually the same as the original chromVAR (averaging
over the noise coming from the random background selection of the
original method).

[^5]: Note, however, that similar motifs will be given similar activity
    estimates, so that it is often hard to know which of a set of
    highly-similar motifs is in fact responsible for the observed
    signal.

The object contains two assays: the `deviations` assay contains the
bias-adjusted deviations from the expectation (by default, the average),
i.e. the difference to the expectation divided by the expectation, and
the `z` assay contains z-scores, i.e. the difference to the expectation
divided by the variance of the expectation.

If your samples/cells have similar library sizes, it is recommended that
you use the `z` assay for downstream analysis (such as differential TF
activity using `limma`). If the samples have very different library
sizes, the `z` scores will be influenced by that, and it might be
preferable to use the `deviations` assay.

The variability of each motif across the dataset is stored in the
rowData of the object :

```{r variability}
rowData(dev)
```

Note that the confidence interval is based on analytic estimates; if bootstrap 
confidence intervals (as in the original approach) are needed, see the 
`computeMotifVariability()` function.

## Individual steps of betterChromVAR

The `betterChromVAR()` function is actually a wrapper around three
steps, which can also be executed individually for more customization,
or to avoid repeating some computation multiple times.

The first step is creating the bias bins and their pairwise sampling
probabilities. This is achieved with:

```{r getBackgroundBins}
bg <- getBackgroundBins(counts)
bg
```

The second step is computing the per-bin background expectations and
variances for each sample or cell:

```{r computeBackgrounds}
bg <- computeBackgrounds(counts, bg)
bg
```

The `bg` object has now been filled with the additional information. It
can now be used for:

```{r computeDeviations}
dev <- computeDeviationsAnalytic(counts, background=bg, annotations=motifMatches)
dev
```

This last call could be made with different annotations, without needing
to recompute the previous steps.

## Including fragment length bias

In addition to the enrichment and GC bias taken into account by the
original `r Biocpkg("chromVAR")`, `r Biocpkg("betterChromVAR")` supports
an optional third dimension: fragment length bias (see
`?getBackgroundBins()` for more information). This requires the
compilation of an additional bias component (the log10-transformed mean
or median fragment length per region). If this is desired, on option is to use 
the counting functions from the
`r Githubpkg("ETHZ-INS/epiwraps", "epiwraps")` package, which can
provide this information.

## Note on single-cell data

If you are using betterChromVAR (or ChromVAR, for that matter) on
single-cell data with multiple cell types, the way to best run it
depends on whether your are interested in differences within cell types,
or between cell types. If interested in differences between cell types,
run it on the entire dataset, specifying the cell types as `grouping`
argument to `betterChromVAR()`. In this way, rare and abundant cell
types will be given the same weight in computing the expectation.

If you are interested in differences within cell types (e.g. between
samples/conditions), you should instead run the method separately for
each cell type. In this way, the background bins will be defined based
on the enrichment in the given cell type, leading to better capture of
the bias. The drawback, however, is that the values won't be comparable
across cell types. To test across conditions, we also recommend using it
on pseudobulk data.

### Execution on Atlas-scale datasets

The only step of the process that needs to be done across the entire
dataset is the computing of the expectation, i.e. the mean (or weighted
mean) for each region across all cells, and the creation of the
background bins, which depends on it. Everything else can easily be
executed in chunks (of cells), as is done for multi-threading in the
`betterChromVAR()` function.

For example, only these functions need to be called on the full dataset
:

```{r getExpectations}
ex <- getExpectation(counts)
bg <- getBackgroundBins(ex, bias=rowData(counts)$bias)
# this is equivalent to bg <- getBackgroundBins(counts)
```

Then we can apply the next steps only on subset of the data:

```{r chunkCompute}
bg2 <- computeBackgrounds(counts[,1:3], bg, expectation = ex)
dev2 <- computeDeviationsAnalytic(counts[,1:3], bg2, motifMatches)
# this should be identical to what we had run on the whole object:
identical(assay(dev)[,1:3], assay(dev2))
```

If data is on disk, rather than in memory (and you want it to stay that
way), make sure to use this procedure. In recent versions,
`getExpectation()` should use chunk-based processing. However it is
quite likely that better performance for this step would be obtained
using for instance using the `aggregateAcrossCells()` function of the
`r Biocpkg("scrapper")` package.

<br/><br/>

# Bias-normalization of bulk ATAC-seq data

The general chromVAR approach, and in particular the deterministic
version implemented here, is also amenable to be used to normalize GC-
and enrichment bias out of bulk ATAC-seq data. This is implemented in
the `CVnorm()` function. In addition, if given a grouping of the samples
(e.g. experimental conditions), the function applies a variance-based
smoothing of the bias correction, inspired by smoothed quantile
normalization (see the `r Biocpkg("qsmooth")` package or
@hicks2018smooth ). In a nutshell, if the bias in a certain background
bin is explained by experimental groups, it will be less corrected than
if it varies across samples of the groups[^6].

[^6]: Note that this is different from the original
    `r Biocpkg("qsmooth")` approach -- see `?CVnorm` for more detail.

Example usage:

```{r CVnorm}
# we assign arbitrary groups to the samples:
counts$group <- rep(LETTERS[1:2], each=5)
# we run the smoothed CVnorm:
counts <- CVnorm(counts, grouping=counts$group)
# (the normal CVnorm could be run by omitting the grouping)
counts
```

This adds a `corrected` assay to the object. Note that although the
assay is corrected for enrichment- and GC-bias (and fragment length bias if
that was provided), it is not corrected for library size differences. Rather, 
it is on the original count scale (although not integer anymore), so that it is 
amenable to use in downstream count-based analysis methods such as 
`r Biocpkg("edgeR")`.

# betterChromVAR with kNN-based background

The [scPRINTER](https://github.com/buenrostrolab/scPrinter) python
package developed by Ruochi Zhang in the Buenrostro lab includes a
GPU-based reimplementation of \textit{chromVAR}, with an important
modification: rather than cutting the bias space into bins, a
continuous, multidimensional bias space space is created from which $k$
nearest neighbors (kNN) are selected as background. This approach scales
nicely to the inclusion of more dimensions of bias. In *scPRINTER*, the 
deviations and variance across the $k$ neighbors is then computed through
permutations, as in the classical \textit{chromVAR} (although sped up by
the GPU). 

If a more complex multi-dimensional bias needs to be taken into account, 
`r Biocpkg("betterChromVAR")` implements a similar algorithm that 
combines our analytical framework with the kNN-based approach. An additional 
feature is that rather than taking the transformed bias space directly to 
compute distances, we weight the different dimensions by their correlation with
peak overdispersion.

The kNN version of betterChromVAR can be used in the following way:

```{r}
bg <- getBackgroundKNN(counts)
dev <- computeDeviationsFromKNN(counts, bg, motifMatches)
```

We note, however, that this approach is considerably slower (although still 
substantially faster than the original chromVAR or the ArchR re-implementation),
requires more memory, and, at the single-cell level, produces slightly more 
variable deviations across replicate cells.
We therefore recommend using the more scalable bin-based variant, unless 
working with bulk data with fragment length bias (or any other extra bias 
dimensions, see the `bias` argument of `getBackgroundKNN()`).

# Appendix

## References

::: {#refs}
:::

## Session info

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{normalizeDevsForSize}
\alias{normalizeDevsForSize}
\title{normalizeDevsForSize}
\usage{
normalizeDevsForSize(dev)
}
\arguments{
\item{dev}{A SummarizedExperiment object as produced by
\code{\link{betterChromVAR}} or \code{\link{computeDeviationsAnalytic}}.}
}
\value{
The \code{dev} object with an additional assay named 'norm'.
}
\description{
Normalizes the z-scores assay of a deviations object to make the scores
comparable across motifs with different number of matches.
}
\examples{
attach(getDummyData())
dev <- betterChromVAR(counts, motifMatches)
dev <- normalizeDevsForSize(dev)
dev
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDeviationsFromKNN.R
\name{computeDeviationsFromKNN}
\alias{computeDeviationsFromKNN}
\title{computeDeviationsFromKNNbg}
\usage{
computeDeviationsFromKNN(
  object,
  cBg,
  annotations,
  l = 1,
  chunkSize = 1000,
  verbose = TRUE
)
}
\arguments{
\item{object}{A SummarizedExperiment or sparse matrix of counts.}

\item{cBg}{The peak-by-peak sparse kNN matrix, as produced by
\code{\link{getBackgroundKNN}}.}

\item{annotations}{Peak annotation (sparse) matrix, with motifs as columns,
or a SummarizedExperiment containing this in the first assay. Values
should be either logical or between 0 and 1.}

\item{l}{Lambda parameter determining the weight by which background peaks
containing the foreground motif are scaled in relative importance. Set to
1 to treat them normally (default), to 0 to exclude them entirely
(potentially unstable, a small value such as \code{0.1} is instead recommended).}

\item{chunkSize}{Number of cells to process simultaneously. Increasing this
will increase speed, but also memory consumption.}

\item{verbose}{Logical; whether to print progress messages.}
}
\value{
A SummarizedExperiment with 'deviations' and 'z' assays. If overall
motif variability and their significance are additionally needed, see
\code{\link{computeMotifVariability}}.
}
\description{
Computes analytical deviations using a nearest-neighbor background matrix
while optionally excluding peaks containing the target motif from its own
background pool.
}
\details{
This method combines the analytic strategy used by \code{betterChromVAR} with
Ruochi Zhang's approach to use a continuous, multidimensional background
space instead of background bins. If \code{l<1} it downweighs (multiplying them
by \code{l}) peaks harboring the tested motif from the corresponding motif's
background.
}
\examples{
attach(getDummyData())
bg <- getBackgroundKNN(counts)
dev <- computeDeviationsFromKNN(object=counts, cBg=bg,
                                annotations=motifMatches)
dev
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDeviationsWeighted.R
\name{computeDeviationsWeighted}
\alias{computeDeviationsWeighted}
\title{computeDeviationsWeighted}
\usage{
computeDeviationsWeighted(
  weightedMotifCounts,
  unweightedPeakCounts,
  annotations,
  bg = NULL,
  retSE = TRUE,
  ...
)
}
\arguments{
\item{weightedMotifCounts}{A matrix of weighted counts per motif (rows) and
sample (columns), or a \code{SummarizedExperiment} containing this as first
assay.}

\item{unweightedPeakCounts}{A matrix of unweighted counts per peak (rows) and
sample (columns), or a \code{SummarizedExperiment} containing this as first
assay.}

\item{annotations}{Peak annotation (sparse) matrix, with motifs as columns,
or a SummarizedExperiment containing this in the first assay. Values
should be either logical or between 0 and 1.}

\item{bg}{Either a \code{\link[betterChromVAR]{bcvBackground-class}} object
as produced by \code{\link{computeBackgrounds}}, or a
\code{SummarizedExperiment} of background peak counts (with bias data in
\code{rowData}). If missing, will be created based on \code{unweightedPeakCounts}.}

\item{retSE}{Logical; whether to return a SummarizedExperiment object.}

\item{...}{Passed to \link{getBackgroundBins} (can for instance be used to
pass bias info if not contained in the objects). Ignored if \code{bg} is a
\code{\link[betterChromVAR]{bcvBackground-class}} object.}
}
\value{
A SummarizedExperiment (or a list if \code{retSE=FALSE}). If overall
motif variability and their significance are additionally needed, see
\code{\link{computeMotifVariability}}.
}
\description{
A variant of \code{\link{computeDeviationsWeighted}} enabling the
computation of deviations from weighted foreground counts. Specifically, this
functions handles the normalization of the difference in magnitude between
the (weighted) foreground and background.
}
\examples{
attach(getDummyData())
# if GC content not already in the object, use:
# counts <- addGCBias(counts, genome=YOUR_GENOME)
# For the purpose of this example, we'll use standard (unweighted counts),
# although at this step we'd compute counts weighted in the desired fashion:
motifCounts <- Matrix::t(motifMatches) \%*\% assay(counts)
dev1 <- computeDeviationsWeighted(motifCounts, counts, motifMatches)
dev1
# in this case, the results are identical to :
dev2 <- betterChromVAR(counts, motifMatches)
stopifnot(identical(assays(dev1), assays(dev2)))
}
\seealso{
\code{\link[=computeDeviationsAnalytic]{computeDeviationsAnalytic()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{addGCBias}
\alias{addGCBias}
\title{addGCBias}
\usage{
addGCBias(object, genome)
}
\arguments{
\item{object}{An object inheriting RangedSummarizedExperiment or GRanges.}

\item{genome}{A BSgenome object or any other genome object supported by
\code{\link[Biostrings]{getSeq}}.}
}
\value{
\code{object} with the GC content in \code{mcols(object)$bias} (if GRanges)
or \code{rowData(object)$bias}.
}
\description{
Add the \code{bias} column to the object's rowData, containing the regions'
proportion of Gs and Cs.
}
\examples{
# not run:
# se <- addGCBias(se, genome)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeBackgrounds.R
\name{computeBackgrounds}
\alias{computeBackgrounds}
\title{computeBackgrounds}
\usage{
computeBackgrounds(
  object,
  bins,
  grouping = NULL,
  expectation = NULL,
  shrinkage = c("none", "average", "smooth"),
  sigma = 1,
  verbose = FALSE
)
}
\arguments{
\item{object}{A SummarizedExperiment (or SingleCellExperiment) with an assay
'counts', or a (sparse) matrix of counts.}

\item{bins}{A \code{bcvBackground} object, as produced by
\code{\link{getBackgroundBins}}.}

\item{grouping}{An optional factor or vector coercible to a factor indicating
the groupings of the columns of \code{object}. This is optionally used to 1)
compute the base expectation such that rare cell types are given as much
weight as abundant ones, and 2) apply shrinkage (if \code{shrinkage!="none"})
on a per-grouping fashion. In single-cell data, the grouping can for
instance be the interaction of samples and cell types. (The name of a
colData column of \code{object} can also be provided.)}

\item{expectation}{Optional vector of length equal to \code{nrow(object)}
giving the expected counts. If NULL, defaults to mean counts (eventually
grouped, see \code{grouping} and \code{\link{getExpectation}}).}

\item{shrinkage}{The method to use to shrink background (i.e. bias) bin
frequencies. Either "average" (shrinks towards the bin's average across
cells/samples of the same group), "smooth" (per-sample 2D smoothing over
the bin matrix, somewhat redundant with \code{w}), or "none" (default).}

\item{sigma}{Sigma parameter for the 2D smoothing. Ignored unless
\code{shrinkage="smooth"}.}

\item{verbose}{Logical; whether to output progress messages.}
}
\value{
A \code{bcvBackground} object with bins*samples slots filled, for use
with \code{\link{computeDeviationsAnalytic}}.
}
\description{
computeBackgrounds
}
\examples{
attach(getDummyData())
# if GC content not already in the object, use:
# counts <- addGCBias(counts, genome=YOUR_GENOME)

# we fist get the background bins:
bg <- getBackgroundBins(counts)
# then we can compute the backgrounds for each sample:
bg <- computeBackgrounds(counts, bg)
# for use in computeDeviationsAnalytic...
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getExpectation}
\alias{getExpectation}
\title{getExpectation}
\usage{
getExpectation(counts, grouping = NULL, normalize = TRUE)
}
\arguments{
\item{counts}{A count matrix, or object inheriting SummarizedExperiment with
a 'counts' assay.}

\item{grouping}{An optional vector of length equal to \code{ncol(counts)},
indicating the grouping of the cells. If provided, cells will be averaged
by group before averaging across groups.}

\item{normalize}{Logical; whether to normalize data between averaging (but
after grouping). Default TRUE and highly recommended if providing
\code{grouping}.}
}
\value{
A vector of expectation for each row of \code{counts}
}
\description{
Computes expected counts (a glorified rowMeans)
}
\examples{
attach(getDummyData())
e <- getExpectation(counts)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getBackgroundBins.R
\name{getBackgroundBins}
\alias{getBackgroundBins}
\title{getBackgroundBins}
\usage{
getBackgroundBins(
  x,
  bias = NULL,
  flbias = NULL,
  w = 0.1,
  bs = NULL,
  pseudo = 0,
  verbose = TRUE
)
}
\arguments{
\item{x}{A SummarizedExperiment containing a 'counts' assay, or a matrix of
counts, or a vector of expected (e.g. mean) counts.}

\item{bias}{A vector of length equal to \code{ncol(object)} specifying the
per-peak bias (i.e. GC content). If omitted, will try to get it
from \code{rowData(x)$bias}.}

\item{flbias}{A vector of length equal to \code{ncol(object)} specifying the
per-peak fragment length bias (e.g. log10-transformed median length of
fragments overlapping each region). If omitted, will try to get it
from \code{rowData(object)$flbias}. If absent, will use 2-dimensional bias bins
(which is already pretty good).}

\item{w}{Standard deviation of the Gaussian kernel.}

\item{bs}{Number of bins per dimension. This can be a single integer (total
bins = \code{bs^2}), or an integer vector of length 2 (if \code{flbias=NULL}) or 3
(in which case there are \code{prod(bs)} total bins). The values specify the
number of bins for, in order: enrichment, GC and fragment length. By
default, \code{bs=50} if \code{flbias} is not provided (mimicking chromVAR), and
\code{bs=c(30, 30, 6)} if it is.}

\item{pseudo}{Optional pseudocount to be added. This should not be needed
with standard workflows.}

\item{verbose}{Whether to print processing info.}
}
\value{
A \code{bcvBackground} object, to be used with
\code{\link{computeBackgrounds}}.
}
\description{
Computes chromVAR-like background (i.e. bias) bins, as well as bin-to-bin
selection probabilities needed for \code{\link{betterChromVAR}}.
}
\details{
The procedure underlying this function is the same as in
\code{chromVAR::getBackgroundPeaks}, with the following differences:
\itemize{
\item Rather than producing a set of background peaks for each input peak, the
function returns peak-to-bin mappings and bin-to-bin background selection
probabilities, which enables an analytic background computation. It is, as
such, entirely deterministic.
\item The function supports the optional use of a third bias dimension, provided
through the \code{flbias} argument, meant for fragment length bias. This is
still an experimental feature.
}
}
\examples{
counts_se <- getDummyData()$counts
background <- getBackgroundBins(counts_se)
}
\references{
Schep A.N., Wu B., Buenrostro J.D., Greenleaf W.J. (2017) chromVAR:
inferring transcription-factor-associated accessibility from
single-cell epigenomic data, Nature Methods, doi: 10.1038/nmeth.4401
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getDummyData}
\alias{getDummyData}
\title{Dummy data for testing purposes}
\usage{
getDummyData(nRegions = 500, nSamples = 10, nMotifs = 5)
}
\arguments{
\item{nRegions}{Number of regions to generate}

\item{nSamples}{Number of samples to generate}

\item{nMotifs}{Number of motifs to generate}
}
\value{
A list with the slots \code{counts} (a peak counts SummarizedExperiment)
and \code{matches} (a sparse matrix of binary motif matches per peaks)
}
\description{
Dummy data for testing purposes
}
\examples{
out <- getDummyData()
(counts <- out$counts)
matches <- out$motifMatches
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeDeviationsAnalytic.R
\name{computeDeviationsAnalytic}
\alias{computeDeviationsAnalytic}
\title{computeDeviationsAnalytic}
\usage{
computeDeviationsAnalytic(
  object,
  background,
  annotations,
  verbose = FALSE,
  retSE = TRUE,
  compute = c("deviations", "z", "variability"),
  denominator = c("global", "local", "none")
)
}
\arguments{
\item{object}{A SummarizedExperiment (or SingleCellExperiment) with an assay
'counts', and with a 'bias' column in \code{rowData(object)}. Note that the
regions should have similar widths.}

\item{background}{A \code{bcvBackground} object with bins*samples slots filled,
as produced by \code{\link{computeBackgrounds}}.}

\item{annotations}{Peak annotation (sparse) matrix, with motifs as columns,
or a SummarizedExperiment containing this in the first assay. Values
should be either logical or between 0 and 1.}

\item{verbose}{Logical; whether to output progress messages.}

\item{retSE}{Logical; whether to return a SummarizedExperiment object.}

\item{compute}{What to compute. Defaults to everything: deviations, z and
motif variability.}

\item{denominator}{The type of denominator to use for the deviations.
Either 'global' (default), i.e. the global expectation (same as the
original chromVAR), 'local' (background expectation of the cell/sample),
or 'none' (denominator of 1). 'global' (default) is recommended.}
}
\value{
A SummarizedExperiment (or a list if \code{retSE=FALSE}).
}
\description{
computeDeviationsAnalytic
}
\examples{
attach(getDummyData())
# if GC content not already in the object, use:
# counts <- addGCBias(counts, genome=YOUR_GENOME)

# we fist get the background bins:
bg <- getBackgroundBins(counts)
# then we compute the backgrounds for each sample:
bg <- computeBackgrounds(counts, bg)
# then we can compute the deviations:
dev <- computeDeviationsAnalytic(counts, bg, motifMatches)
dev
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betterChromVAR.R
\name{betterChromVAR}
\alias{betterChromVAR}
\title{betterChromVAR}
\usage{
betterChromVAR(
  object,
  annotations,
  grouping = NULL,
  nthreads = NULL,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{A SummarizedExperiment (or SingleCellExperiment) with an assay
'counts', and with a 'bias' column in \code{rowData(object)}. Note that the
regions should have similar widths.}

\item{annotations}{Peak annotation (sparse) matrix, with motifs as columns,
or a SummarizedExperiment containing this in the first assay. Values
should be either logical or between 0 and 1.}

\item{grouping}{An optional factor or vector coercible to a factor indicating
the groupings of the columns of \code{object}. This is optionally used to
compute the base expectation such that rare cell types are given as much
weight as abundant ones. In single-cell data, the grouping can for
instance be the interaction of samples and cell types. This should either
be a vector coercible to factor of length equal to \code{ncol(object)}, or a
character of length 1 specifying a column of \code{colData(object)}.}

\item{nthreads}{Either an integer scalar indicating the number of threads to
use, or a \code{BiocParallelParam} object.}

\item{verbose}{Logical; whether to output progress messages (default FALSE).}

\item{...}{Passed to \code{\link{getBackgroundBins}}.}
}
\value{
A SummarizedExperiment containing the adjusted deviations and
z-scores for each motif/sample. The rowData additionally contains the
number of motif matches and their variability.
}
\description{
A fast, analytic implementation of \code{chromVAR}.
This is a wrapper around the \code{\link{getBackgroundBins}},
\code{\link{computeBackgrounds}}, and \code{\link{computeDeviationsAnalytic}}
steps. It additionally allows for multithreading. For more control or
optimization, see the individual steps.
}
\details{
Contrarily to the original chromVAR, this function is entirely deterministic,
and achieves higher precision and much higher efficiency through two changes:
\enumerate{
\item working with expected background sampling mean and variances, rather than
actual permutations, and 2) computing expectations and variance at the level
of bias bins, instead of in the peak-space. The function additionally
includes experimental bias shrinkage options, the possibility to handle
annotations that are not binary (e.g. probability scores) and a third
bias dimension (fragment length bias, which should be stored in
\code{rowData(object)$flbias} see \code{\link{getBackgroundBins}} for details).
}
}
\examples{
attach(getDummyData())
# if GC content not already in the object, use:
# counts <- addGCBias(counts, genome=YOUR_GENOME)
dev <- betterChromVAR(counts, motifMatches)
dev
# note that this is the exact equivalent of doing:
# bg <- getBackgroundBins(counts)
# bg <- computeBackgrounds(counts, bg)
# dev <- computeDeviationsAnalytic(counts, bg, motifMatches)
}
\references{
Schep A.N., Wu B., Buenrostro J.D., Greenleaf W.J. (2017) chromVAR:
inferring transcription-factor-associated accessibility from
single-cell epigenomic data, Nature Methods, doi: 10.1038/nmeth.4401
}
\author{
Pierre-Luc Germain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcvBackground.R
\name{as.list,bcvBackground-method}
\alias{as.list,bcvBackground-method}
\alias{show,bcvBackground-method}
\alias{[,bcvBackground,ANY,ANY,ANY-method}
\title{Coerce bcvBackground to a list}
\usage{
\S4method{as.list}{bcvBackground}(x)

\S4method{show}{bcvBackground}(object)

\S4method{[}{bcvBackground,ANY,ANY,ANY}(x, i, j, ..., drop = TRUE)
}
\arguments{
\item{x}{A \code{bcvBackground} object.}

\item{object}{A \code{bcvBackground} object.}

\item{i, j}{Indices for subsetting (if j is provided, i is ignored).}

\item{...}{Additional arguments.}

\item{drop}{Logical, whether to drop dimensions.}
}
\value{
A \code{list} containing the slots of the object.

Nothing, prints an overview of the object.

An \code{bcvBackground} object.
}
\description{
Coerce bcvBackground to a list

Show a bcvBackground object

Subsetting a bcvBackground
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computeVariability.R
\name{computeMotifVariability}
\alias{computeMotifVariability}
\title{computeMotifVariability}
\usage{
computeMotifVariability(
  z,
  confInt = 0.95,
  n = 100,
  method = c("bonett", "normal", "bootstrap")
)
}
\arguments{
\item{z}{A matrix of z-scores (with motifs as rows), or a
\code{SummaziedExperiment} object with such a matrix in assay \code{z} (as produced
e.g. by \code{\link{computeDeviationsAnalytic}}).}

\item{confInt}{The the confidence interval (a numeric scalar between 0 and
1, default 0.95).}

\item{n}{The number of bootstrap samples. Ignored unless
\code{method="bootstrap"}.}

\item{method}{The method used to compute the confidence interval. 'normal'
computes it analytically, assuming that the z-scores are normally
distributed. 'bonett' (default) adjusts this analytic estimate for
kurtosis (based on Bonett, Computational Statistics & Data Analysis, 2006).
'bootstrap' uses bootstrapping, which does not scale very well.}
}
\value{
A data.frame containing the variability and confidence interval
around it, as well as significance, for each motif. If \code{z} is a
\code{SummarizedExperiment}, the data.frame will be stored in \code{rowData(z)}.
}
\description{
computeMotifVariability
}
\examples{
# we generate random z-scores:
z <- matrix(rnorm(mean=rnorm(10), sd=runif(10, max=2), 200), nrow=10)
var <- computeMotifVariability
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getBackgroundKNN.R
\name{getBackgroundKNN}
\alias{getBackgroundKNN}
\title{getBackgroundKNN}
\usage{
getBackgroundKNN(
  se,
  expectation = NULL,
  bias = NULL,
  k = 50,
  weights = c("linear", "poly", "none"),
  pseudo = 0.1,
  ...
)
}
\arguments{
\item{se}{A SummarizedExperiment containing a 'counts' assay, or a matrix of
counts.}

\item{expectation}{A vector of expectations. If NULL, will use the mean
counts of \code{se}}

\item{bias}{A data.frame of sources of bias (beside \code{expectation}) to
consider (one per column), with the same nrow as \code{se}.}

\item{k}{Number of nearest neighbors to use.}

\item{weights}{How to weigh the different bias dimensions. If "none", they
will not be re-weighted. If 'linear' (default), they are weighted by the
absolute Pearson correlation with the over-dispersion. If "poly", by the
R^2 of a 2nd degree polynomial fit of the over-dispersion.}

\item{pseudo}{Pseudocount for log transformation.}

\item{...}{Passed to \code{\link[BiocNeighbors]{findKNN}}.}
}
\value{
A sparse peak-by-peak kNN matrix.
}
\description{
Computes k-nearest neighbors for each peaks based on the continuous,
multidimensional bias space. This is inpsired by Ruochi Zhang's approach in
scPrinter.
}
\examples{
SE <- getDummyData()$counts
bg <- getBackgroundKNN(SE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{shrinkColumnProps}
\alias{shrinkColumnProps}
\title{shrinkColumnProps}
\usage{
shrinkColumnProps(x, shrinkTo = NULL, var.theo = FALSE)
}
\arguments{
\item{x}{A matrix of counts, with features as rows and samples as columns.}

\item{shrinkTo}{A vector (of length equal to \code{nrow(x)}) of sampling
probabilities to shrink towards, or a matrix (of the same dimensions as \code{x})
of such probabilities. If omitted, will be the weighted mean of the columns'
relative frequencies.}

\item{var.theo}{Logical; whether to use theoretical (i.e. binomial) variances
of the proportions, rather than the observed (weighted) variance.}
}
\value{
A matrix of the same dimensions as \code{x} representing the shrunk
column-wise proportions.
}
\description{
Empirical Bayes shrinkage of a matrix of counts towards a prior proportion
(by default the mean across columns).
}
\examples{
# generate a matrix of 5 sampling (with different total counts) of 20 
# features based on the same base frequency :
baseFreq <- abs(rnorm(20))
baseFreq <- baseFreq/sum(baseFreq)
mat <- sapply(c(10,20,30,40,50), function(tot){
  rpois(length(baseFreq), baseFreq*tot)
})
# apply shrinkage and confirm that shrunk proportions are better correlated
shrunk_mat <- shrinkColumnProps(mat)
mean(cor(shrunk_mat))>mean(cor(mat))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CVnorm.R
\name{CVnorm}
\alias{CVnorm}
\title{CVnorm: chromVAR-inspired ATAC-seq normalization}
\usage{
CVnorm(
  object,
  bias = NULL,
  grouping = NULL,
  smoothGrouping = grouping,
  shrinkMode = c("dampen", "qsmooth"),
  toAssay = "corrected",
  bs = NULL,
  w = 0.1,
  Z = FALSE,
  useWidthAdj = NULL,
  enforceZeros = TRUE
)
}
\arguments{
\item{object}{A matrix of counts, or a SummarizedExperiment-like object with
an assay named 'counts'.}

\item{bias}{A vector of length equal to \code{ncol(object)} specifying the
per-peak bias (i.e. GC content). If omitted, will try to get it
from \code{rowData(object)$bias}.}

\item{grouping}{Optional grouping for the baseline expectation (prevents
bias toward more abundant groups). This should either be a vector coercible
to factor of length equal to \code{ncol(object)}, of a character of length 1
specifying a column of \code{colData(object)} (if \code{object} is a
\code{SummarizedExperiment}).}

\item{smoothGrouping}{Optional grouping to determine correction strength.
If bias is consistent within these groups, correction is reduced. Accepts
the same type of inputs as \code{grouping}, and by default takes the same
values.}

\item{shrinkMode}{The way to perform the group-based shrinkage. With
\code{shrinkMode="dampen"} (default), no corrected is applied in bins when the
bias is entirely explained by groups. \code{shrinkMode="qsmooth"} instead
reproduces the logic of the \code{qsmooth} package: if variance in bias is
chiefly explained by groups, between group bias will not be corrected, but
within-group differences will be. Using this prior to differential analysis
however leads to increase Type I error rate, and it should therefore not be
used for downstream application.}

\item{toAssay}{The name of the assay in which to store the corrected data
(default 'corrected'). Ignored unless \code{object} is a
SummarizedExperiment-like object.}

\item{bs}{Number of bins per dimension (see \code{\link{getBackgroundBins}}).}

\item{w}{Standard deviation of the Gaussian kernel for bin smoothing.}

\item{Z}{Logical; whether to return standardized residuals (Z-scores)
instead of the (default) corrected counts.}

\item{useWidthAdj}{Whether to adjust for the different width of the regions.
If omitted, will be TRUE if the average absolute difference to the median
width is greater than 10\% of the median width. If TRUE, will adjust for
\code{pmax(200L,width)}. If an integer scalar, will adjust for
\code{pmax(useWidthAdj,width)}.}

\item{enforceZeros}{Logical; whether to enforce that zero counts should
remain zeroes after correction (ignored if \code{Z=TRUE}).}
}
\value{
If \code{object} is a matrix, then a matrix of corrected counts of the
same dimensions. If \code{object} is a SummarizedExperiment-like object, then
the object is returned with an extra assay named based on \code{toAssay}.
}
\description{
Corrects ATAC peak counts by removing the effects of technical biases
(GC/accessibility) using the chromVAR background binning approach and an
optional variance-based bias shrinkage (inspired from the qsmooth package)
to preserve group biological signal.
}
\examples{
counts_se <- getDummyData()$counts
# if GC content not already in the object, use:
# counts_se <- addGCBias(counts_se, genome=YOUR_GENOME)
counts_se <- CVnorm(counts_se)
}
\references{
\itemize{
\item Schep A.N., Wu B., Buenrostro J.D., Greenleaf W.J. (2017) chromVAR:
inferring transcription-factor-associated accessibility from
single-cell epigenomic data, Nature Methods, doi: 10.1038/nmeth.4401
\item Hicks SC, Okrah K, Paulson JN, Quackenbush J, Irizarry RA, Corrado
Bravo H (2018). “Smooth quantile normalization.” Biostatistics 19 (2),
doi: 10.1093/biostatistics/kxx028
}
}
\author{
Pierre-Luc Germain
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bcvBackground.R
\docType{class}
\name{bcvBackground-class}
\alias{bcvBackground-class}
\title{Bin and background data for betterChromVAR (for internal use)}
\description{
Bin and background data for betterChromVAR (for internal use)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getBackgroundBins.R
\name{sampleBackgroundPeaks}
\alias{sampleBackgroundPeaks}
\title{sampleBackgroundPeaks}
\usage{
sampleBackgroundPeaks(background, niterations = 50)
}
\arguments{
\item{background}{A \code{bcvBackground} produced by
\code{\link{getBackgroundBins}}.}

\item{niterations}{Number of background peaks to sample for each target peak.}
}
\value{
A peaks x niterations matrix of integers representing the indices of
the sampled background peaks.
}
\description{
Given a background generated by \code{\link{getBackgroundBins}}, samples
background peaks for each input peak.
}
\details{
This function is not used by \code{\link{betterChromVAR}}, which is
deterministic, but for other applications requiring an outputs similar to
that of the original \code{\link[chromVAR]{getBackgroundPeaks}}.
}
\examples{
counts_se <- getDummyData()$counts
background <- getBackgroundBins(counts_se)
bg_peaks <- sampleBackgroundPeaks(background, niterations=20)
}
